from enum import Enum
from typing import (
    Optional,
)

from pydantic import BaseModel, Field


class Provider(str, Enum):
    ai21 = "ai21"
    anthropic = "anthropic"
    azure = "azure"
    cerebras = "cerebras"
    cohere = "cohere"
    databricks = "databricks"
    deepinfra = "deepinfra"
    fireworks = "fireworks"
    gemini = "gemini"
    groq = "groq"
    lepton = "lepton"
    moonshot = "moonshot"
    octoai = "octoai"
    openai = "openai"
    openrouter = "openrouter"
    perplexity = "perplexity"
    together = "together"
    vertex = "vertex"
    bigmodel = "bigmodel"
    tensoropera = "tensoropera"
    xai = "xai"
    deepseek = "deepseek"
    lambdalabs = "lambdalabs"


class Vendor(str, Enum):
    nvidia = "nvidia"
    meta_llama = "meta-llama"
    perplexity = "perplexity"
    databricks = "databricks"
    mistralai = "mistralai"
    mixtral = "mixtral"
    pygmalionai = "pygmalionai"
    qwen = "qwen"
    Qwen = "Qwen"  # For Deepseek
    openai = "openai"
    anthropic = "anthropic"
    cohere = "cohere"
    gemini = "gemini"
    together = "together"
    nous_hermes = "nous-hermes"
    hermes_llama = "hermes-llama"
    moonshot = "moonshot"
    wizardlm = "wizardlm"
    google = "google"
    ai21 = "ai21"
    O1 = "01"
    starcoder = "starcoder"
    gryphe = "gryphe"
    microsoft = "microsoft"
    moescape = "moescape"
    bigmodel = "bigmodel"
    deepseek_ai = "deepseek-ai"
    grok = "grok"
    liquid = "liquid"
    deepseek = "deepseek"


class ConversationRole(str, Enum):
    user = "user"
    assistant = "assistant"
    system = "system"
    function = "function"
    tool = "tool"
    developer = "developer"


class AdapterFinishReason(str, Enum):
    stop = "stop"
    length = "length"
    tool_calls = "tool_calls"
    content_filter = "content_filter"
    function_call = "function_call"


class ContentType(str, Enum):
    text = "text"
    image_url = "image_url"


class Cost(BaseModel):
    prompt: float
    completion: float
    request: float = 0.0


class ModelProperties(BaseModel):
    open_source: bool = False
    chinese: bool = False
    gdpr_compliant: bool = False
    is_nsfw: bool = False


class Model(BaseModel):
    test_async: bool = True

    name: str
    api_name: Optional[str] = None
    vendor_name: str
    provider_name: str
    cost: Cost
    context_length: int
    completion_length: Optional[int] = None

    is_response: bool = False

    supports_chat: bool = True
    supports_completion: bool = True
    supports_functions: bool = True

    supports_streaming: bool = True
    supports_vision: bool = True
    supports_n: bool = True
    supports_tools: bool = True
    supports_tools_streaming: bool = False
    supports_tools_choice: bool = True
    supports_tools_choice_required: bool = True
    supports_json_output: bool = True
    supports_json_content: bool = True
    supports_stop: bool = True
    supports_max_tokens: bool = True
    supports_max_completion_tokens: bool = True

    can_user: bool = True
    can_vision_multiple: bool = True
    can_repeating_roles: bool = True
    can_empty_content: bool = True
    can_temperature: bool = True
    can_presence_penalty: bool = True
    can_repetition_penalty: bool = True
    can_top_p: bool = True
    can_top_k: bool = True
    can_min_p: bool = True

    can_system: bool = True
    can_system_only: bool = True
    can_system_multiple: bool = True
    can_system_last: bool = True

    can_assistant_only: bool = True
    can_assistant_first: bool = True
    can_assistant_last: bool = True

    properties: ModelProperties = Field(default_factory=ModelProperties)

    def get_path(self) -> str:
        return f"{self.provider_name}/{self.vendor_name}/{self.name}"

    def get_api_path(self) -> str:
        return self.api_name if self.api_name else self.name

    def __str__(self) -> str:
        return self.get_path()


__all__ = [
    "Provider",
    "Vendor",
    "ConversationRole",
    "AdapterFinishReason",
    "ContentType",
    "Cost",
    "ModelProperties",
    "Model",
]



from abc import abstractmethod
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Generator,
    Generic,
    Iterable,
    Literal,
    Optional,
    TypeVar,
    Unpack,
    cast,
    overload,
)
from llm_adapters.abstract_adapters.api_key_adapter_mixin import ApiKeyAdapterMixin
from llm_adapters.abstract_adapters.adapter import Adapter
from llm_adapters.abstract_adapters.provider_adapter_mixin import ProviderAdapterMixin
from llm_adapters.client_cache import client_cache
from llm_adapters.constants import OVERRIDE_ALL_BASE_URLS
from llm_adapters.general_utils import (
    EMPTY_CONTENT,
    delete_none_values,
    stream_generator_auto_close,
)
from llm_adapters.types import (
    AdapterChatCompletion,
    AdapterChatCompletionChunk,
    AdapterCompletion,
    AdapterCompletionChunk,
    AdapterException,
    AdapterStreamAsyncChatCompletion,
    AdapterStreamAsyncCompletion,
    AdapterStreamSyncChatCompletion,
    AdapterStreamSyncCompletion,
    ContentType,
    ConversationRole,
    Model,
    ModelProperties,
    ChatCompletionCreateArgs,
    CompletionCreateArgs,
    NOT_GIVEN,
    NotGiven,
    ChatCompletionMessageParam,
    ChatCompletionUserMessageParam,
)

CLIENT_SYNC = TypeVar("CLIENT_SYNC")
CLIENT_ASYNC = TypeVar("CLIENT_ASYNC")


class SDKChatAdapter(
    Adapter,
    ApiKeyAdapterMixin,
    ProviderAdapterMixin,
    Generic[CLIENT_SYNC, CLIENT_ASYNC],
):
    _client_sync: CLIENT_SYNC
    _client_async: CLIENT_ASYNC

    def __init__(
        self,
    ) -> None:
        super().__init__()
        self._setup_clients(self.get_api_key())

    def _get_or_create_client(
        self, api_key: str, client_type: Literal["sync", "async"]
    ) -> Any:
        base_url = OVERRIDE_ALL_BASE_URLS or self.get_base_sdk_url()
        client = client_cache.get_client(base_url, api_key, client_type)
        if not client:
            client = getattr(self, f"_create_client_{client_type}")(
                api_key=api_key, base_url=base_url
            )
            client_cache.set_client(base_url, api_key, client_type, client)
        return client

    def _setup_clients(self, api_key: str) -> None:
        self._client_sync = self._get_or_create_client(api_key, "sync")
        self._client_async = self._get_or_create_client(api_key, "async")

    @abstractmethod
    def _call_sync(self) -> Callable[..., Any]:
        pass

    @abstractmethod
    def _call_async(self) -> Callable[..., Any]:
        pass

    @abstractmethod
    def _call_completion_sync(self) -> Callable[..., Any]:
        pass

    @abstractmethod
    def _call_completion_async(self) -> Callable[..., Any]:
        pass

    @abstractmethod
    def _create_client_sync(self, base_url: str, api_key: str) -> CLIENT_SYNC:
        pass

    @abstractmethod
    def _create_client_async(self, base_url: str, api_key: str) -> CLIENT_ASYNC:
        pass

    @abstractmethod
    def get_base_sdk_url(self) -> str:
        pass

    @abstractmethod
    def _extract_response(self, request: Any, response: Any) -> AdapterChatCompletion:
        pass

    @abstractmethod
    def _extract_completion_response(
        self, request: Any, response: Any
    ) -> AdapterCompletion:
        pass

    @abstractmethod
    def _extract_stream_response(
        self, request: Any, response: Any, state: dict[str, Any]
    ) -> AdapterChatCompletionChunk:
        pass

    @abstractmethod
    def _extract_completion_stream_response(
        self, request: Any, response: Any, state: dict[str, Any]
    ) -> AdapterCompletionChunk:
        pass

    def _adjust_temperature(self, temperature: float) -> float:
        return temperature

    def _verify(self, **kwargs: Unpack[ChatCompletionCreateArgs]) -> None:
        if not self.get_model().supports_tools and "tools" in kwargs:
            raise AdapterException(f"Tools is not supported on {self.get_model().name}")

        if (
            not self.get_model().supports_tools_choice
            and kwargs.get("tool_choices") is not None
        ):
            raise AdapterException(
                f"Tool choice is not supported on {self.get_model().name}"
            )
        if (kwargs.get("max_tokens") or 0) > (self.get_model().completion_length or 0):
            raise AdapterException(
                f"max_tokens {kwargs.get('max_tokens')} should be less than max completion length {self.get_model().completion_length} for {self.get_model().name}"
            )

        if not self.get_model().supports_streaming and kwargs.get("stream") is True:
            raise AdapterException(
                f"Streaming is not supported on {self.get_model().name}"
            )

        if not self.get_model().supports_functions and "functions" in kwargs:
            raise AdapterException(
                f"Function calling is not supported on {self.get_model().name}"
            )

        if self.get_model().supports_n is False and (kwargs.get("n") or 0) >= 1:
            raise AdapterException(f"n is not supported on {self.get_model().name}")

        if not self.get_model().supports_json_content:
            for message in kwargs["messages"]:
                if not isinstance(message["content"], str):
                    raise AdapterException(
                        f"JSON content is not supported on {self.get_model().name}"
                    )

        if not self.get_model().supports_vision:
            for message in kwargs["messages"]:
                if not isinstance(message["content"], str) and message["content"]:
                    for content in message["content"]:
                        if (
                            not isinstance(content, str)
                            and content["type"] == ContentType.image_url.value
                        ):
                            raise AdapterException(
                                f"Vision is not supported on {self.get_model().name}"
                            )

        if not self.get_model().supports_json_output and (
            cast(dict[str, Any], kwargs.get("response_format") or {}).get("type")
            == "json_schema"
            or cast(dict[str, Any], kwargs.get("response_format") or {}).get("type")
            == "json_object"
        ):
            raise AdapterException(
                f"JSON response format is not supported on {self.get_model().name}"
            )

    # TODO: Check if a "system" message is between two "user" messages
    def _format_messages(
        self, **kwargs: Unpack[ChatCompletionCreateArgs]
    ) -> Iterable[ChatCompletionMessageParam]:
        messages = list(kwargs["messages"])

        # Check if messages are empty
        if len(messages) == 0:
            messages.append(
                ChatCompletionUserMessageParam(
                    role=ConversationRole.user.value, content=EMPTY_CONTENT
                )
            )

        # Convert empty string to EMPTY_CONTENT if not supported
        if not self.get_model().can_empty_content:
            for message in messages:
                if (
                    isinstance(message["content"], str)
                    and message["content"].strip() == ""
                ):
                    message["content"] = EMPTY_CONTENT
                elif isinstance(message["content"], list):
                    for content in message["content"]:
                        if (
                            isinstance(content, dict)
                            and "text" in content
                            and content["text"].strip() == ""
                        ):
                            content["text"] = EMPTY_CONTENT

        # Add empty user message if assistant only
        if (
            not self.get_model().can_assistant_only
            and messages[0]["role"] == ConversationRole.assistant.value
        ):
            messages.append(
                ChatCompletionUserMessageParam(
                    role=ConversationRole.user.value, content=EMPTY_CONTENT
                )
            )

        # Covered by can_system_last
        # Add empty user message if system only
        if (
            not self.get_model().can_system_only
            and messages[0]["role"] == ConversationRole.system.value
        ):
            messages.append(
                ChatCompletionUserMessageParam(
                    role=ConversationRole.user.value, content=EMPTY_CONTENT
                )
            )

        # Change system prompt roles to user if not supported
        if not self.get_model().can_system_multiple:
            for messageId, message in enumerate(messages[1:], start=1):
                if message["role"] == ConversationRole.system.value:
                    messages[messageId] = ChatCompletionUserMessageParam(
                        role=ConversationRole.user.value, content=message["content"]
                    )

        # Change system prompt roles to user
        if not self.get_model().can_system:
            for messageId, message in enumerate(messages):
                if message["role"] == ConversationRole.system.value:
                    messages[messageId] = ChatCompletionUserMessageParam(
                        role=ConversationRole.user.value, content=message["content"]
                    )

        # If the first message is assistant, insert an empty user message
        if (
            self.get_model().can_assistant_first is False
            and messages[0]["role"] == ConversationRole.assistant.value
        ):
            messages.insert(
                0,
                ChatCompletionUserMessageParam(
                    role=ConversationRole.user.value, content=EMPTY_CONTENT
                ),
            )

        # If the last message is assistant, add an empty user message
        if (
            self.get_model().can_assistant_last is False
            and messages[-1]["role"] == ConversationRole.assistant.value
        ):
            messages.append(
                ChatCompletionUserMessageParam(
                    role=ConversationRole.user.value, content=EMPTY_CONTENT
                )
            )

        # If the last message is system, add an empty user message
        if (
            self.get_model().can_system_last is False
            and messages[-1]["role"] == ConversationRole.system.value
        ):
            messages.append(
                ChatCompletionUserMessageParam(
                    role=ConversationRole.user.value, content=EMPTY_CONTENT
                )
            )

        # Join messages from the same role
        if not self.get_model().can_repeating_roles:
            result: list[ChatCompletionMessageParam] = []
            grouped_message: Optional[ChatCompletionMessageParam] = None

            for message in messages:
                if grouped_message and message["role"] == grouped_message["role"]:
                    if isinstance(grouped_message["content"], str) and isinstance(
                        message["content"], str
                    ):
                        grouped_message["content"] = (
                            f"{grouped_message['content']}\n{message['content']}"
                        )
                    elif isinstance(grouped_message["content"], list) and isinstance(
                        message["content"], list
                    ):
                        grouped_message["content"].extend(message["content"])
                    elif isinstance(grouped_message["content"], list) and isinstance(
                        message["content"], str
                    ):
                        grouped_message["content"].append(
                            {"type": ContentType.text.value, "text": message["content"]}
                        )
                    elif isinstance(grouped_message["content"], str) and isinstance(
                        message["content"], list
                    ):
                        grouped_message["content"] = [
                            {
                                "type": ContentType.text.value,
                                "text": grouped_message["content"],
                            }
                        ]  # type: ignore
                        grouped_message["content"].extend(message["content"])  # type: ignore
                else:
                    if grouped_message:
                        result.append(grouped_message)
                    grouped_message = message

            if grouped_message:
                result.append(grouped_message)

            messages = result

        return messages

    # TODO: add support for n removal
    def _get_params(
        self,
        *,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> dict[str, Any]:
        self._verify(**kwargs)
        messages = self._format_messages(**kwargs)

        # List of attribute flags and their corresponding kwargs
        attributes = [
            ("can_user", "user"),
            ("can_min_p", "min_p"),
            ("can_top_k", "top_k"),
            ("can_top_p", "top_p"),
            ("can_presence_penalty", "presence_penalty"),
            ("can_repetition_penalty", "repetition_penalty"),
            ("can_temperature", "temperature"),
        ]

        for attr_flag, kwarg_key in attributes:
            if not getattr(self.get_model(), attr_flag, False) and kwarg_key in kwargs:
                kwargs[kwarg_key] = None  # type: ignore

        temperature = kwargs.get("temperature")

        params: dict[str, Any] = {
            "model": self.get_model().get_api_path(),
            "messages": messages,
            **({"stream": stream} if stream is not NOT_GIVEN else {}),
            **(
                {"temperature": self._adjust_temperature(temperature)}
                if temperature is not None
                else {}
            ),
        }

        for top_param in [
            "model",
            "messages",
            "stream",
            "temperature",
            "max_tokens",
            "extra_body",
            "extra_query",
            "extra_headers",
        ]:
            if top_param in kwargs:
                if top_param not in params:
                    params[top_param] = kwargs[top_param]  # type: ignore
                del kwargs[top_param]  # type: ignore

        params["extra_body"] = kwargs

        return delete_none_values(params)

    def get_model(self) -> Model:
        if self._current_model is None:
            raise ValueError("Model is not set")
        return self._current_model

    def get_model_properteis(self, model_name: str) -> ModelProperties:
        for model in self.get_supported_models():
            if model.name == model_name:
                return model.properties
        raise ValueError(f"Model {model_name} not found")

    def set_api_key(self, api_key: str) -> None:
        super().set_api_key(api_key)
        self._setup_clients(api_key)

    @overload
    def execute_completion_sync(
        self,
        *,
        stream: Literal[True],
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterStreamSyncCompletion: ...
    @overload
    def execute_completion_sync(
        self,
        *,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion: ...
    def execute_completion_sync(
        self,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion | AdapterStreamSyncCompletion:
        kwargs.pop("model", None)

        response = self._call_completion_sync()(
            model=self.get_model().get_api_path(), stream=stream, **kwargs
        )

        if not stream:
            return self._extract_completion_response(
                request=kwargs["prompt"], response=response
            )

        def stream_response() -> Generator[AdapterCompletionChunk, Any, None]:
            state: dict[str, Any] = {}
            try:
                for chunk in response:
                    yield self._extract_completion_stream_response(
                        request=kwargs["prompt"], response=chunk, state=state
                    )
            except Exception as e:
                raise AdapterException(f"Error in streaming response: {e}") from e
            finally:
                response.close()

        return AdapterStreamSyncCompletion(response=stream_response())

    @overload
    async def execute_completion_async(
        self,
        stream: Literal[True],
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterStreamAsyncCompletion: ...
    @overload
    async def execute_completion_async(
        self,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion: ...
    async def execute_completion_async(
        self,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion | AdapterStreamAsyncCompletion:
        kwargs.pop("model", None)

        response = await self._call_completion_async()(
            model=self.get_model().get_api_path(), stream=stream, **kwargs
        )

        if not stream:
            return self._extract_completion_response(
                request=kwargs["prompt"], response=response
            )

        async def stream_response() -> AsyncGenerator[AdapterCompletionChunk, None]:
            state: dict[str, Any] = {}
            try:
                async for chunk in response:
                    yield self._extract_completion_stream_response(
                        request=kwargs["prompt"], response=chunk, state=state
                    )
            except Exception as e:
                raise AdapterException(f"Error in streaming response: {e}") from e
            finally:
                await response.close()

        return AdapterStreamAsyncCompletion(response=stream_response())

    @overload
    def execute_chat_completion_sync(
        self,
        stream: Literal[True],
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterStreamSyncChatCompletion: ...
    @overload
    def execute_chat_completion_sync(
        self,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion: ...
    def execute_chat_completion_sync(
        self,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion | AdapterStreamSyncChatCompletion:
        response = self._call_sync()(**self._get_params(stream=stream, **kwargs))

        if not stream:
            return self._extract_response(request=kwargs, response=response)

        def stream_response() -> Generator[AdapterChatCompletionChunk, Any, None]:
            state: dict[str, Any] = {}
            try:
                for chunk in response:
                    yield self._extract_stream_response(
                        request=kwargs, response=chunk, state=state
                    )
            except Exception as e:
                raise AdapterException(f"Error in streaming response: {e}") from e
            finally:
                if hasattr(response, "close"):
                    response.close()

        return AdapterStreamSyncChatCompletion(response=stream_response())

    @overload
    async def execute_chat_completion_async(
        self,
        stream: Literal[True],
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterStreamAsyncChatCompletion: ...
    @overload
    async def execute_chat_completion_async(
        self,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion: ...
    async def execute_chat_completion_async(
        self,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion | AdapterStreamAsyncChatCompletion:
        response = await self._call_async()(**self._get_params(stream=stream, **kwargs))

        if not stream:
            return self._extract_response(request=kwargs, response=response)

        async def stream_response() -> AsyncGenerator[AdapterChatCompletionChunk, None]:
            state: dict[str, Any] = {}
            async with stream_generator_auto_close(response):
                try:
                    async for chunk in response:
                        yield self._extract_stream_response(
                            request=kwargs, response=chunk, state=state
                        )
                except Exception as e:
                    raise AdapterException(f"Error in streaming response: {e}") from e
                finally:
                    if hasattr(response, "close"):
                        await response.close()

        return AdapterStreamAsyncChatCompletion(response=stream_response())



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import Model


class CustomOpenAISDKChatProviderAdapter(OpenAISDKChatAdapter):
    def __init__(self, base_url: str) -> None:
        self.base_url = base_url
        super().__init__()

    @staticmethod
    def get_supported_models() -> list[Model]:
        return []

    def get_base_sdk_url(self) -> str:
        return self.base_url

    @staticmethod
    def get_api_key_name() -> str:
        return ""



# from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
# from llm_adapters.types import Model, ModelProperties, Provider, Vendor


# class MoonshotModel(Model):
#     provider_name: str = Provider.moonshot.value
#     vendor_name: str = Vendor.moonshot.value

#     supports_completion: bool = False
#     supports_vision: bool = False
#     supports_tools: bool = False
#     supports_n: bool = False
#     supports_max_completion_tokens: bool = False
#     supports_max_tokens: bool = False

#     can_empty_content: bool = False
#     can_min_p: bool = False
#     can_top_p: bool = False

#     properties: ModelProperties = ModelProperties(chinese=True)


# # Cost measured in CNY, converted to USD on Apr 27 2024
# MODELS: list[Model] = [
#     # MoonshotModel(
#     #     name="moonshot-v1-8k",
#     #     cost=Cost(prompt=1.66e-6, completion=1.66e-6),
#     #     context_length=8000,
#     # ),
#     # MoonshotModel(
#     #     name="moonshot-v1-32k",
#     #     cost=Cost(prompt=3.32e-6, completion=3.32e-6),
#     #     context_length=32000,
#     # ),
#     # MoonshotModel(
#     #     name="moonshot-v1-128k",
#     #     cost=Cost(prompt=8.29e-6, completion=8.29e-6),
#     #     context_length=128000,
#     # ),
# ]


# class MoonshotSDKChatProviderAdapter(OpenAISDKChatAdapter):
#     @staticmethod
#     def get_supported_models() -> list[Model]:
#         return MODELS

#     @staticmethod
#     def get_api_key_name() -> str:
#         return "MOONSHOT_API_KEY"

#     def get_base_sdk_url(self) -> str:
#         return "https://api.moonshot.cn/v1"



from llm_adapters.types.openai import (
    NOT_GIVEN,
    NotGiven,
    Completion,
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
    Query,
    Body,
    Headers,
    Metadata,
    ChatModel,
    ChatCompletionAudioParam,
    FunctionCall,
    FunctionCreate,
    FunctionToolCall,
    ChatCompletionModality,
    ChatCompletionPredictionContentParam,
    ResponseFormat,
    ChatCompletionReasoningEffort,
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
    ChatCompletionMessageToolCall,
    Choice,
    ChatCompletionMessage,
    CompletionUsage,
    ChatCompletionUserMessageParam,
    ChoiceChunk,
    ChoiceDelta,
    FilePurpose,
    Batch,
    HttpxBinaryResponseContent,
    SyncCursorPage,
    AsyncCursorPage,
    AsyncPaginator,
)

from llm_adapters.types.request import (
    ChatCompletionCreateKwargs,
    CompletionCreateKwargs,
    ChatCompletionCreateArgs,
    CompletionCreateArgs,
)

from llm_adapters.types.exception import (
    AdapterException,
    AdapterModelNotFoundException,
    AdapterRateLimitException,
)

from llm_adapters.types.model import (
    Provider,
    Vendor,
    ConversationRole,
    AdapterFinishReason,
    ContentType,
    Cost,
    ModelProperties,
    Model,
)

from llm_adapters.types.response import (
    AdapterChatCompletion,
    AdapterStreamChatCompletion,
    AdapterStreamSyncChatCompletion,
    AdapterStreamAsyncChatCompletion,
    AdapterChatCompletionChunk,
    AdapterCompletion,
    AdapterCompletionChunk,
    AdapterStreamCompletion,
    AdapterStreamSyncCompletion,
    AdapterStreamAsyncCompletion,
    # Deprecated
    Turn,
    FunctionOutputTurn,
    ToolOutputTurn,
    ToolsCallTurn,
    ImageDetailsType,
    VisionImageDetails,
    TextContentEntry,
    ImageContentEntry,
    ContentTurn,
    TurnType,
    Conversation,
    Prompt,
    AdapterResponse,
    AdapterStreamResponse,
)

from llm_adapters.types.files import (
    FileObject,
)

__all__ = [
    "NotGiven",
    "NOT_GIVEN",
    "Completion",
    "ChatCompletion",
    "ChatCompletionChunk",
    "ChatCompletionMessageParam",
    "Query",
    "Body",
    "Headers",
    "ChatModel",
    "ChatCompletionAudioParam",
    "FunctionCall",
    "FunctionCreate",
    "ChatCompletionModality",
    "ChatCompletionPredictionContentParam",
    "ResponseFormat",
    "ChatCompletionReasoningEffort",
    "ChatCompletionStreamOptionsParam",
    "ChatCompletionToolChoiceOptionParam",
    "ChatCompletionToolParam",
    "ChatCompletionCreateKwargs",
    "CompletionCreateKwargs",
    "Completion",
    "ChatCompletionChunk",
    "AdapterException",
    "AdapterRateLimitException",
    "Provider",
    "Vendor",
    "ConversationRole",
    "AdapterFinishReason",
    "ContentType",
    "Cost",
    "ModelProperties",
    "Model",
    "AdapterChatCompletion",
    "AdapterStreamChatCompletion",
    "AdapterStreamSyncChatCompletion",
    "AdapterStreamAsyncChatCompletion",
    "AdapterChatCompletionChunk",
    "AdapterCompletion",
    "AdapterCompletionChunk",
    "AdapterStreamCompletion",
    "AdapterStreamSyncCompletion",
    "AdapterStreamAsyncCompletion",
    "AdapterModelNotFoundException",
    "ChatCompletionCreateArgs",
    "CompletionCreateArgs",
    "ChatCompletionUserMessageParam",
    "ChoiceChunk",
    "ChoiceDelta",
    "ChatCompletionMessageToolCall",
    "Choice",
    "ChatCompletionMessage",
    "CompletionUsage",
    "FunctionToolCall",
    "FileObject",
    "FilePurpose",
    "Metadata",
    "Batch",
    "HttpxBinaryResponseContent",
    "SyncCursorPage",
    "AsyncCursorPage",
    "AsyncPaginator",
    # Deprecated
    "Turn",
    "FunctionOutputTurn",
    "ToolOutputTurn",
    "ToolsCallTurn",
    "ImageDetailsType",
    "VisionImageDetails",
    "TextContentEntry",
    "ImageContentEntry",
    "ContentTurn",
    "TurnType",
    "Conversation",
    "Prompt",
    "AdapterResponse",
    "AdapterStreamResponse",
]



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import (
    Cost,
    Model,
    ModelProperties,
    Provider,
    Vendor,
)


class GeminiModel(Model):
    provider_name: str = Provider.gemini.value
    vendor_name: str = Vendor.gemini.value

    supports_completion: bool = False
    supports_tools: bool = False

    can_empty_content: bool = False
    can_system_last: bool = False

    properties: ModelProperties = ModelProperties(gdpr_compliant=True)


# TODO: setup pricing for longer contexts
MODELS: list[Model] = [
    GeminiModel(
        name="gemini-1.5-pro",
        cost=Cost(prompt=1.25e-6, completion=5.00e-6),
        context_length=2097152,
        completion_length=8192,
    ),
    GeminiModel(
        name="gemini-1.5-flash-8b",
        cost=Cost(prompt=0.0375e-6, completion=0.15e-6),
        context_length=1048576,
        completion_length=8192,
    ),
    GeminiModel(
        name="gemini-1.5-flash",
        cost=Cost(prompt=0.075e-6, completion=0.30e-6),
        context_length=1048576,
        completion_length=8192,
    ),
    GeminiModel(
        name="gemini-2.0-flash",
        cost=Cost(prompt=0.10e-6, completion=0.40e-6),
        context_length=1048576,
        completion_length=8192,
    ),
    GeminiModel(
        name="gemini-2.0-flash-lite",
        cost=Cost(prompt=0.075e-6, completion=0.30e-6),
        context_length=1048576,
        completion_length=8192,
        can_assistant_last=False,
    ),
    GeminiModel(
        name="gemini-2.0-flash",
        cost=Cost(prompt=0.0375e-6, completion=0.15e-6),
        context_length=1048576,
        completion_length=8192,
    ),
]


class GeminiSDKChatProviderAdapter(OpenAISDKChatAdapter):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "GEMINI_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://generativelanguage.googleapis.com/v1beta/openai"



# from enum import Enum
# from typing import (
#     Any,
#     Dict,
#     Iterable,
#     List,
#     Literal,
#     Optional,
#     Union,
# )
# from pydantic import BaseModel
# from llm_adapters.types import (
#     AdapterStreamChatCompletion,
#     AdapterChatCompletion,
#     ConversationRole,
#     ChatCompletionMessageToolCall,
#     ChatCompletionMessageParam,
# )
# from openai.types.chat.chat_completion_message import FunctionCall


# class Turn(BaseModel, use_enum_values=True):
#     role: Union[ConversationRole]
#     content: str


# class FunctionOutputTurn(BaseModel, use_enum_values=True):
#     role: Literal[ConversationRole.function] = ConversationRole.function
#     content: Optional[str] = None
#     name: str


# class ToolOutputTurn(BaseModel, use_enum_values=True):
#     role: Literal[ConversationRole.tool] = ConversationRole.tool
#     content: Optional[str] = None
#     tool_call_id: str


# class FunctionCallTurn(BaseModel, use_enum_values=True):
#     role: Literal[ConversationRole.assistant] = ConversationRole.assistant
#     function_call: FunctionCall
#     content: None = None


# class ToolsCallTurn(BaseModel, use_enum_values=True):
#     role: Literal[ConversationRole.assistant] = ConversationRole.assistant
#     tool_calls: list[ChatCompletionMessageToolCall]
#     content: None = None


# class ImageDetailsType(str, Enum):
#     high = "high"
#     low = "low"
#     auto = "auto"


# class ContentType(str, Enum):
#     text = "text"
#     image_url = "image_url"


# class VisionImageDetails(BaseModel, use_enum_values=True):
#     url: str
#     details: Optional[ImageDetailsType] = ImageDetailsType.auto


# class TextContentEntry(BaseModel, use_enum_values=True):
# #     type: Literal[ContentType.text] = ContentType.text
#     text: str


# class ImageContentEntry(BaseModel, use_enum_values=True):
# #     type: Literal[ContentType.image_url] = ContentType.image_url
#     image_url: VisionImageDetails


# class ContentTurn(BaseModel, use_enum_values=True, validate_assignment=True):
#     role: str = ConversationRole.user
#     content: List[Union[ImageContentEntry, TextContentEntry]]


# TurnType = Union[
#     Turn,
#     FunctionOutputTurn,
#     ToolOutputTurn,
#     ToolsCallTurn,
#     ContentTurn,
#     FunctionCallTurn,
# ]


# class Conversation(BaseModel):
#     turns: List[TurnType]

#     def __init__(
#         self,
#         turns: Union[
#             "Conversation",
#             List[TurnType],
#             Dict[str, List[TurnType]],
#         ],
#     ):
#         if isinstance(turns, Conversation):
#             turns = turns.turns
#         elif isinstance(turns, dict) and "turns" in turns:
#             turns = turns["turns"]
#         super().__init__(turns=turns)

#     def __getitem__(self, index: int) -> TurnType:
#         return self.turns[index]

#     def __setitem__(self, index: int, value: TurnType) -> None:
#         if not isinstance(value, Turn):
#             raise ValueError("Value must be an instance of Turn")
#         self.turns[index] = value

#     def __len__(self) -> int:
#         return len(self.turns)

#     def __iter__(self) -> Any:
#         return iter(self.turns)

#     def is_last_turn_vision_query(self) -> bool:
#         if len(self.turns):
#             contentTurn = self.turns[len(self.turns) - 1]
#         else:
#             return False

#         if isinstance(contentTurn, ContentTurn):
#             return any(
#                 content.type == ContentType.image_url for content in contentTurn.content
#             )

#         return False

#     def convert_to_prompt(self) -> "Prompt":
#         return Prompt("".join([f"{turn.role}: {turn.content}" for turn in self.turns]))

#     def convert_to_openai_format(self) -> Iterable[ChatCompletionMessageParam]:
#         return self.model_dump()["turns"]  # type: ignore


# # Deprecated
# class Prompt(str):
#     def convert_to_conversation(self) -> Conversation:
#         return Conversation(turns=[Turn(role=ConversationRole.user, content=self)])


# # Deprecated, Use AdapterChatCompletion
# AdapterResponse = AdapterChatCompletion

# # Deprecated, Use AdapterStreamChatCompletion
# AdapterStreamResponse = AdapterStreamChatCompletion

# __all__ = [
#     "Turn",
#     "FunctionOutputTurn",
#     "ToolOutputTurn",
#     "ToolsCallTurn",
#     "ImageDetailsType",
#     "ContentType",
#     "VisionImageDetails",
#     "TextContentEntry",
#     "ImageContentEntry",
#     "ContentTurn",
#     "TurnType",
#     "Conversation",
#     "Prompt",
#     "AdapterResponse",
#     "AdapterStreamResponse",
# ]



# from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
# from llm_adapters.types import Cost, Model, Provider, Vendor


# class LeptonModel(Model):
#     provider_name: str = Provider.lepton.value

#     supports_vision: bool = False
#     supports_json_content: bool = False

#     can_assistant_first: bool = False
#     can_assistant_last: bool = False
#     can_assistant_only: bool = False

#     can_system_multiple: bool = False
#     can_repeating_roles: bool = False

#     can_min_p: bool = False
#     can_top_k: bool = False


# MODELS: list[Model] = [
#     LeptonModel(
#         name="mistral-7b",
#         cost=Cost(prompt=0.07e-6, completion=0.07e-6),
#         context_length=8192,
#         vendor_name=Vendor.mistralai.value,
#     ),
#     LeptonModel(
#         name="mixtral-8x7b",
#         cost=Cost(prompt=0.50e-6, completion=0.50e-6),
#         context_length=32768,
#         vendor_name=Vendor.mistralai.value,
#     ),
#     LeptonModel(
#         name="qwen2-72b",
#         cost=Cost(prompt=0.8e-6, completion=0.8e-6),
#         context_length=128000,
#         vendor_name=Vendor.qwen.value,
#     ),
#     LeptonModel(
#         name="wizardlm-2-7b",
#         cost=Cost(prompt=0.07e-6, completion=0.07e-6),
#         context_length=32000,
#         vendor_name=Vendor.wizardlm.value,
#     ),
#     LeptonModel(
#         name="wizardlm-2-8x22b",
#         cost=Cost(prompt=1.0e-6, completion=1.0e-6),
#         context_length=64000,
#         vendor_name=Vendor.wizardlm.value,
#         supports_tools_choice_required=False,
#     ),
#     LeptonModel(
#         name="dolphin-mixtral-8x7b",
#         cost=Cost(prompt=0.5e-6, completion=0.5e-6),
#         context_length=32000,
#         vendor_name=Vendor.mistralai.value,
#     ),
# ]


# class LeptonSDKChatProviderAdapter(OpenAISDKChatAdapter):
#     @staticmethod
#     def get_supported_models() -> list[Model]:
#         return MODELS

#     @staticmethod
#     def get_api_key_name() -> str:
#         return "LEPTON_API_KEY"

#     def get_base_sdk_url(self) -> str:
#         return f"https://{self.get_model().name if self._current_model else ''}.lepton.run/api/v1/"

#     def _set_current_model(self, model: Model) -> None:
#         super()._set_current_model(model)

#         self._setup_clients(self.get_api_key())



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import Cost, Model, Provider, Vendor


class DeepInfraModel(Model):
    provider_name: str = Provider.deepinfra.value

    supports_completion: bool = False
    supports_vision: bool = False
    supports_tools: bool = False
    supports_json_content: bool = False
    supports_n: bool = False

    can_assistant_only: bool = False
    can_system_only: bool = False
    can_min_p: bool = False
    can_top_k: bool = False

    def get_api_path(self) -> str:
        return f"{self.vendor_name}/{self.name}"


# TODO: check completion length
# TODO: try lower case model names
# TODO: check json support
# TODO: check function/tools calling support
MODELS: list[Model] = [
    DeepInfraModel(
        name="Llama-3.3-70B-Instruct",
        cost=Cost(prompt=0.23e-6, completion=0.40e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Llama-3.3-70B-Instruct-Turbo",
        cost=Cost(prompt=0.12e-6, completion=0.30e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="phi-4",
        cost=Cost(prompt=0.07e-6, completion=0.14e-6),
        context_length=16384,
        vendor_name=Vendor.microsoft,
    ),
    DeepInfraModel(
        name="DeepSeek-V3",
        cost=Cost(prompt=0.85e-6, completion=0.90e-6),
        context_length=16000,
        vendor_name=Vendor.deepseek_ai.value,
        can_system_only=False,
    ),
    DeepInfraModel(
        name="DeepSeek-R1",
        cost=Cost(prompt=0.85e-6, completion=2.50e-6),
        context_length=16000,
        vendor_name=Vendor.deepseek_ai.value,
        can_system_only=False,
    ),
    DeepInfraModel(
        name="DeepSeek-R1-Distill-Llama-70B",
        cost=Cost(prompt=0.23e-6, completion=0.69e-6),
        context_length=131072,
        vendor_name=Vendor.deepseek_ai.value,
        can_system_only=False,
    ),
    DeepInfraModel(
        name="QwQ-32B-Preview",
        cost=Cost(prompt=0.12e-6, completion=0.18e-6),
        context_length=32768,
        vendor_name=Vendor.Qwen,
    ),
    DeepInfraModel(
        name="Qwen2.5-Coder-32B-Instruct",
        cost=Cost(prompt=0.07e-6, completion=0.16e-6),
        context_length=32768,
        vendor_name=Vendor.Qwen,
    ),
    DeepInfraModel(
        name="Qwen2.5-72B-Instruct",
        cost=Cost(prompt=0.23e-6, completion=0.40e-6),
        context_length=32768,
        vendor_name=Vendor.Qwen,
    ),
    DeepInfraModel(
        name="WizardLM-2-8x22B",
        cost=Cost(prompt=0.50e-6, completion=0.50e-6),
        context_length=65536,
        vendor_name=Vendor.microsoft,
    ),
    DeepInfraModel(
        name="Llama-4-Maverick-17B-128E-Instruct-FP8",
        cost=Cost(prompt=20e-6, completion=60e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
        can_empty_content=False,
    ),
    DeepInfraModel(
        name="Llama-4-Scout-17B-16E-Instruct",
        cost=Cost(prompt=10e-6, completion=30e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
        can_empty_content=False,
    ),
    DeepInfraModel(
        name="DeepSeek-R1-Turbo",
        cost=Cost(prompt=100e-6, completion=300e-6),
        context_length=32768,
        vendor_name=Vendor.deepseek_ai,
    ),
    DeepInfraModel(
        name="QwQ-32B",
        cost=Cost(prompt=15e-6, completion=20e-6),
        context_length=131072,
        vendor_name=Vendor.Qwen,
    ),
    DeepInfraModel(
        name="DeepSeek-V3-0324",
        cost=Cost(prompt=40e-6, completion=89e-6),
        context_length=163840,
        vendor_name=Vendor.deepseek_ai,
    ),
    DeepInfraModel(
        name="gemma-3-27b-it",
        cost=Cost(prompt=10e-6, completion=20e-6),
        context_length=131072,
        vendor_name=Vendor.google,
        can_empty_content=False,
    ),
    DeepInfraModel(
        name="gemma-3-12b-it",
        cost=Cost(prompt=5e-6, completion=10e-6),
        context_length=131072,
        vendor_name=Vendor.google,
        can_empty_content=False,
    ),
    DeepInfraModel(
        name="gemma-3-4b-it",
        cost=Cost(prompt=2e-6, completion=4e-6),
        context_length=131072,
        vendor_name=Vendor.google,
        can_empty_content=False,
    ),
    DeepInfraModel(
        name="Phi-4-multimodal-instruct",
        cost=Cost(prompt=5e-6, completion=10e-6),
        context_length=131072,
        vendor_name=Vendor.microsoft,
    ),
    DeepInfraModel(
        name="DeepSeek-R1-Distill-Qwen-32B",
        cost=Cost(prompt=12e-6, completion=18e-6),
        context_length=131072,
        vendor_name=Vendor.deepseek_ai,
    ),
    DeepInfraModel(
        name="Meta-Llama-3.1-405B-Instruct",
        cost=Cost(prompt=80e-6, completion=80e-6),
        context_length=32768,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Meta-Llama-3.1-70B-Instruct",
        cost=Cost(prompt=23e-6, completion=40e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Meta-Llama-3.1-8B-Instruct",
        cost=Cost(prompt=3e-6, completion=5e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Meta-Llama-3.1-70B-Instruct-Turbo",
        cost=Cost(prompt=12e-6, completion=30e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Meta-Llama-3.1-8B-Instruct-Turbo",
        cost=Cost(prompt=2e-6, completion=5e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Mistral-Small-24B-Instruct-2501",
        cost=Cost(prompt=7e-6, completion=14e-6),
        context_length=32768,
        vendor_name=Vendor.mistralai,
    ),
    DeepInfraModel(
        name="Mixtral-8x7B-Instruct-v0.1",
        cost=Cost(prompt=24e-6, completion=24e-6),
        context_length=32768,
        vendor_name=Vendor.mistralai,
    ),
    DeepInfraModel(
        name="Llama-3.2-90B-Vision-Instruct",
        cost=Cost(prompt=35e-6, completion=40e-6),
        context_length=32768,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Llama-3.2-11B-Vision-Instruct",
        cost=Cost(prompt=5.5e-6, completion=5.5e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Llama-3.2-1B-Instruct",
        cost=Cost(prompt=1e-6, completion=1e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
    DeepInfraModel(
        name="Llama-3.2-3B-Instruct",
        cost=Cost(prompt=1.5e-6, completion=2.5e-6),
        context_length=131072,
        vendor_name=Vendor.meta_llama,
    ),
]


class DeepInfraSDKChatProviderAdapter(OpenAISDKChatAdapter):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "DEEPINFRA_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.deepinfra.com/v1/openai"

    def _adjust_temperature(self, temperature: float) -> float:
        return temperature / 2



from .ai21_sdk_chat_provider_adapter import AI21SDKChatProviderAdapter
from .anthropic_sdk_chat_provider_adapter import AnthropicSDKChatProviderAdapter
from .cerebras_sdk_chat_provider_adapter import CerebrasSDKChatProviderAdapter
from .cohere_sdk_chat_provider_adapter import CohereSDKChatProviderAdapter
from .deepinfra_sdk_chat_provider_adapter import DeepInfraSDKChatProviderAdapter
from .openai_sdk_chat_provider_adapter import OpenAISDKChatProviderAdapter
from .gemini_sdk_chat_provider_adapter import GeminiSDKChatProviderAdapter
from .together_sdk_chat_provider_adapter import TogetherSDKChatProviderAdapter
from .deepseek_sdk_chat_provider_adapter import DeepSeekSDKChatProviderAdapter
from .lambdalabs_sdk_chat_provider_adapter import LambdaLabsSDKChatProviderAdapter

# from .xai_sdk_chat_provider_adapter import XAISDKChatProviderAdapter

# from .azure_sdk_chat_provider_adapter import AzureSDKChatProviderAdapter
# from .databricks_sdk_chat_provider_adapter import DatabricksSDKChatProviderAdapter
# from .deepinfra_sdk_chat_provider_adapter import DeepInfraSDKChatProviderAdapter
# from .fireworks_sdk_chat_provider_adapter import FireworksSDKChatProviderAdapter
# from .moescape_sdk_chat_provider_adapter import MoescapeSDKChatProviderAdapter
# from .tensoropera_sdk_chat_provider_adapter import TensorOperaSDKChatProviderAdapter
# from .groq_sdk_chat_provider_adapter import GroqSDKChatProviderAdapter
# from .lepton_sdk_chat_provider_adapter import LeptonSDKChatProviderAdapter
# from .moonshot_sdk_chat_provider_adapter import MoonshotSDKChatProviderAdapter
# from .octoai_sdk_chat_provider_adapter import OctoaiSDKChatProviderAdapter
# from .openrouter_sdk_chat_provider_adapter import OpenRouterSDKChatProviderAdapter
# from .perplexity_sdk_chat_provider_adapter import PerplexitySDKChatProviderAdapter
# from .bigmodel_provider_adapter import BigModelSDKChatProviderAdapter

__all__ = [
    "AI21SDKChatProviderAdapter",
    "AnthropicSDKChatProviderAdapter",
    "CerebrasSDKChatProviderAdapter",
    "CohereSDKChatProviderAdapter",
    "OpenAISDKChatProviderAdapter",
    "DeepInfraSDKChatProviderAdapter",
    "GeminiSDKChatProviderAdapter",
    "TogetherSDKChatProviderAdapter",
    "DeepSeekSDKChatProviderAdapter",
    "LambdaLabsSDKChatProviderAdapter",
    # # "AzureSDKChatProviderAdapter",
    # # "DatabricksSDKChatProviderAdapter",
    # "DeepInfraSDKChatProviderAdapter",
    # "FireworksSDKChatProviderAdapter",
    # "MoescapeSDKChatProviderAdapter",
    # "TensorOperaSDKChatProviderAdapter",
    # "GroqSDKChatProviderAdapter",
    # "LeptonSDKChatProviderAdapter",
    # "MoonshotSDKChatProviderAdapter",
    # "OctoaiSDKChatProviderAdapter",
    # "OpenRouterSDKChatProviderAdapter",
    # "PerplexitySDKChatProviderAdapter",
    # "BigModelSDKChatProviderAdapter",
]



from enum import Enum
import json
import time
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    TypedDict,
    Union,
    Unpack,
    cast,
)

from anthropic import Anthropic, AsyncAnthropic
from anthropic.types import (
    Message,
    RawContentBlockDeltaEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStreamEvent,
)
from anthropic.types.message_create_params import (
    Metadata,
    ToolChoice,
    ToolChoiceToolChoiceAny,
    ToolChoiceToolChoiceAuto,
    ToolChoiceToolChoiceTool,
)
from anthropic.types.message_param import MessageParam
from anthropic.types.text_delta import TextDelta
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.tool_param import ToolParam
from httpx import AsyncClient, Client, Limits, Timeout

from openai.types.chat.chat_completion_chunk import Choice as ChoiceChunk, ChoiceDelta

from llm_adapters.abstract_adapters.sdk_chat_adapter import SDKChatAdapter
from llm_adapters.constants import (
    HTTP_CONNECT_TIMEOUT,
    HTTP_TIMEOUT,
    MAX_CONNECTIONS_PER_PROCESS,
    MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS,
)
from llm_adapters.general_utils import delete_none_values, process_image_url_anthropic
from llm_adapters.types import (
    AdapterChatCompletion,
    AdapterChatCompletionChunk,
    AdapterFinishReason,
    ConversationRole,
    Cost,
    Model,
    Provider,
    Vendor,
    ChatCompletionMessageToolCall,
    FunctionToolCall,
    Choice,
    ChatCompletionMessage,
    CompletionUsage,
    NotGiven,
    ChatCompletionCreateArgs,
)

CACHE_PROMPT_TOKEN_PREMIUM = 0.2
CACHED_PROMPT_TOKEN_DISCOUNT = 0.1


class AnthropicModel(Model):
    vendor_name: str = Vendor.anthropic.value
    provider_name: str = Provider.anthropic.value

    # TODO: fix vision
    supports_vision: bool = False
    # TODO: fix tools
    supports_tools: bool = False
    # TODO: fix functions
    supports_functions: bool = False

    supports_completion: bool = False
    supports_n: bool = False

    can_system: bool = False
    can_empty_content: bool = False


MODELS: list[Model] = [
    AnthropicModel(
        name="claude-3-opus-latest",
        cost=Cost(prompt=15.00e-6, completion=75.00e-6),
        context_length=200000,
        completion_length=4096,
    ),
    AnthropicModel(
        name="claude-3-opus-20240229",
        cost=Cost(prompt=15.00e-6, completion=75.00e-6),
        context_length=200000,
        completion_length=4096,
    ),
    AnthropicModel(
        name="claude-3-5-haiku-latest",
        cost=Cost(prompt=0.80e-6, completion=4.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
    AnthropicModel(
        name="claude-3-5-haiku-20241022",
        cost=Cost(prompt=0.80e-6, completion=4.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
    AnthropicModel(
        name="claude-3-5-sonnet-latest",
        cost=Cost(prompt=3.00e-6, completion=15.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
    AnthropicModel(
        name="claude-3-5-sonnet-20241022",
        cost=Cost(prompt=3.00e-6, completion=15.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
    AnthropicModel(
        name="claude-3-5-sonnet-20240620",
        cost=Cost(prompt=3.00e-6, completion=15.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
    AnthropicModel(
        name="claude-3-7-sonnet-20250219",
        cost=Cost(prompt=3.00e-6, completion=15.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
    AnthropicModel(
        name="claude-3-7-sonnet-latest",
        cost=Cost(prompt=3.00e-6, completion=15.00e-6),
        context_length=200000,
        completion_length=8192,
    ),
]


class AnthropicFinishReason(str, Enum):
    end_turn = "end_turn"
    max_tokens = "max_tokens"
    stop_sequence = "stop_sequence"
    tool_use = "tool_use"


FINISH_REASON_MAPPING: Dict[AnthropicFinishReason, AdapterFinishReason] = {
    AnthropicFinishReason.end_turn: AdapterFinishReason.stop,
    AnthropicFinishReason.max_tokens: AdapterFinishReason.length,
    AnthropicFinishReason.stop_sequence: AdapterFinishReason.stop,
    AnthropicFinishReason.tool_use: AdapterFinishReason.tool_calls,
}


class AnthropicCreate(TypedDict, total=False):
    model: str
    max_tokens: int
    messages: Iterable[MessageParam]
    metadata: Optional[Metadata]
    stop_sequences: Optional[List[str]]
    stream: Optional[Literal[False] | Literal[True]]
    system: Optional[Union[str, Iterable[TextBlockParam]]]
    temperature: Optional[float]
    tool_choice: Optional[ToolChoice]
    tools: Optional[Iterable[ToolParam]]
    top_k: Optional[int]
    top_p: Optional[float]


class AnthropicSDKChatProviderAdapter(SDKChatAdapter[Anthropic, AsyncAnthropic]):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "ANTHROPIC_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.anthropic.com"

    def _call_completion_sync(self) -> Callable[..., Any]:
        raise NotImplementedError

    def _call_completion_async(self) -> Callable[..., Any]:
        raise NotImplementedError

    def _call_sync(self) -> Callable[..., Any]:
        return self._client_sync.messages.create

    def _call_async(self) -> Callable[..., Any]:
        return self._client_async.messages.create

    def _create_client_sync(self, base_url: str, api_key: str) -> Anthropic:
        return Anthropic(
            base_url=base_url,
            api_key=api_key,
            max_retries=0,
            http_client=Client(
                limits=Limits(
                    max_connections=MAX_CONNECTIONS_PER_PROCESS,
                    max_keepalive_connections=MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS,
                ),
                timeout=Timeout(timeout=HTTP_TIMEOUT, connect=HTTP_CONNECT_TIMEOUT),
            ),
        )

    def _create_client_async(self, base_url: str, api_key: str) -> AsyncAnthropic:
        return AsyncAnthropic(
            base_url=base_url,
            api_key=api_key,
            max_retries=0,
            http_client=AsyncClient(
                limits=Limits(
                    max_connections=MAX_CONNECTIONS_PER_PROCESS,
                    max_keepalive_connections=MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS,
                ),
                timeout=Timeout(timeout=HTTP_TIMEOUT, connect=HTTP_CONNECT_TIMEOUT),
            ),
        )

    def _adjust_temperature(self, temperature: float) -> float:
        return temperature / 2

    def _get_params(
        self,
        *,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> dict[str, Any]:
        openai_messages = list(kwargs["messages"])
        kwargs["messages"] = openai_messages

        system_prompt: Optional[Union[str, Iterable[TextBlockParam]]] = None

        # Extract system prompt if it's the first message, only works for str
        if (
            len(openai_messages)
            and openai_messages[0]["role"] == ConversationRole.system.value
            and isinstance(openai_messages[0]["content"], str)
        ):
            system_prompt = openai_messages[0]["content"]
            kwargs["messages"] = openai_messages[1:]

        params = super()._get_params(stream=stream, **kwargs)

        anthropic_messages = params["messages"]

        # Remove trailing whitespace from the last assistant message
        if (
            len(anthropic_messages)
            and anthropic_messages[-1]["role"] == ConversationRole.assistant.value
        ):
            if isinstance(anthropic_messages[-1]["content"], str):
                anthropic_messages[-1]["content"] = anthropic_messages[-1][
                    "content"
                ].rstrip()
            elif anthropic_messages[-1]["content"]:
                messages_content_list = list(anthropic_messages[-1]["content"])
                if messages_content_list[-1]["type"] == "text":
                    messages_content_list[-1]["text"] = messages_content_list[-1][
                        "text"
                    ].rstrip()
                anthropic_messages[-1]["content"] = messages_content_list

        # Include base64-encoded images in the request
        for message in anthropic_messages:
            if (
                isinstance(message["content"], list)
                and message["role"] == ConversationRole.user.value
            ):
                anthropic_content = []

                for content in message["content"]:
                    if content["type"] == "text":
                        anthropic_content.append(content)
                    elif content["type"] == "image_url":
                        anthropic_content.append(
                            process_image_url_anthropic(content["image_url"]["url"])
                        )

                message["content"] = anthropic_content

        # Convert tools to anthropic format
        openai_tools = params.get("tools")
        openai_tools_choice = params.get("tool_choice")

        anthropic_tools: Optional[list[ToolParam]] = None

        if openai_tools:
            anthropic_tools = []
            for openai_tool in openai_tools:
                anthropic_tool = ToolParam(
                    name=openai_tool["function"]["name"],
                    description=openai_tool["function"]["description"],
                    input_schema={
                        "type": openai_tool["function"]["parameters"]["type"],
                        "properties": openai_tool["function"]["parameters"][
                            "properties"
                        ],
                        "required": openai_tool["function"]["parameters"]["required"],
                    },
                )

                anthropic_tools.append(anthropic_tool)

        anthropic_tool_choice: Optional[ToolChoice] = None

        if openai_tools_choice == "required":
            anthropic_tool_choice = ToolChoiceToolChoiceAny(type="any")
        elif openai_tools_choice == "auto":
            anthropic_tool_choice = ToolChoiceToolChoiceAuto(type="auto")
        elif openai_tools_choice == "none":
            anthropic_tools = None
        elif isinstance(openai_tools_choice, dict):
            anthropic_tool_choice = ToolChoiceToolChoiceTool(
                name=openai_tools_choice["function"]["name"],
                type="tool",
            )

        anthropic_create: AnthropicCreate = {
            "model": self.get_model().get_api_path(),
            "max_tokens": params.get("max_tokens", self.get_model().completion_length),
            "messages": anthropic_messages,
            "metadata": params.get("metadata"),
            "stop_sequences": params.get("stop_sequences"),
            "stream": bool(params.get("stream")),
            "system": system_prompt,
            "temperature": params.get("temperature"),
            "tool_choice": anthropic_tool_choice,
            "tools": anthropic_tools,
            "top_k": params.get("top_k"),
            "top_p": params.get("top_p"),
        }

        return delete_none_values(cast(dict[str, Any], anthropic_create))

    def _extract_response(
        self, request: Any, response: Message
    ) -> AdapterChatCompletion:
        finish_reason = FINISH_REASON_MAPPING.get(
            AnthropicFinishReason(response.stop_reason), AdapterFinishReason.stop
        )

        choices: list[Choice] = []
        for content in response.content:
            if content.type == "text":
                choices.append(
                    Choice(
                        index=len(choices),
                        finish_reason=finish_reason.value,
                        message=ChatCompletionMessage(
                            role=ConversationRole.assistant.value,
                            content=content.text,
                        ),
                    )
                )
            elif content.type == "tool_use":
                choices.append(
                    Choice(
                        index=len(choices),
                        finish_reason=finish_reason.value,
                        message=ChatCompletionMessage(
                            role=ConversationRole.assistant.value,
                            tool_calls=[
                                ChatCompletionMessageToolCall(
                                    id=content.id,
                                    type="function",
                                    function=FunctionToolCall(
                                        name=content.name,
                                        arguments=json.dumps(content.input),
                                    ),
                                )
                            ],
                        ),
                    )
                )
            else:
                raise ValueError("Unsupported response")

        cached_creation_input_tokens = (
            response.usage.cache_creation_input_tokens
            if response.usage.cache_creation_input_tokens
            else 0
        )
        cached_read_input_tokens = (
            response.usage.cache_read_input_tokens
            if response.usage.cache_read_input_tokens
            else 0
        )

        cost = (
            cached_creation_input_tokens * CACHE_PROMPT_TOKEN_PREMIUM
            + cached_read_input_tokens * CACHED_PROMPT_TOKEN_DISCOUNT
            + self.get_model().cost.prompt
            * (response.usage.input_tokens - cached_read_input_tokens)
            + self.get_model().cost.completion * response.usage.output_tokens
            + self.get_model().cost.request
        )

        usage = CompletionUsage(
            prompt_tokens=response.usage.input_tokens,
            completion_tokens=response.usage.output_tokens,
            total_tokens=response.usage.input_tokens + response.usage.output_tokens,
        )

        return AdapterChatCompletion(
            id=response.id,
            created=int(time.time()),
            model=self.get_model().name,
            object="chat.completion",
            cost=cost,
            usage=usage,
            choices=choices,
        )

    def _extract_stream_response(
        self, request: Any, response: RawMessageStreamEvent, state: dict[str, Any]
    ) -> AdapterChatCompletionChunk:
        choice_chunk = ChoiceChunk(
            index=0,
            delta=ChoiceDelta(role=ConversationRole.assistant.value, content=""),
        )

        if isinstance(response, RawMessageStartEvent):
            state["id"] = response.message.id
            state["created"] = int(time.time())
        elif isinstance(response, RawContentBlockDeltaEvent) and isinstance(
            response.delta, TextDelta
        ):
            choice_chunk.delta.content = response.delta.text
        elif isinstance(response, RawMessageDeltaEvent) and response.delta.stop_reason:
            choice_chunk.finish_reason = FINISH_REASON_MAPPING.get(
                AnthropicFinishReason(response.delta.stop_reason),
                AdapterFinishReason.stop,
            ).value

        return AdapterChatCompletionChunk(
            id=state["id"],
            choices=[choice_chunk],
            created=state["created"],
            model=self.get_model().name,
            object="chat.completion.chunk",
        )

    def _extract_completion_response(
        self,
        request: Any,
        response: Any,
    ) -> Any:
        raise NotImplementedError

    def _extract_completion_stream_response(
        self, request: Any, response: Any, state: dict[str, Any]
    ) -> Any:
        raise NotImplementedError



# from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
# from llm_adapters.types import Cost, Model, ModelProperties, Provider, Vendor


# class PerplexityModel(Model):
#     provider_name: str = Provider.perplexity.value

#     supports_completion: bool = False
#     supports_last_system: bool = False
#     supports_json_output: bool = False
#     supports_tools: bool = False
#     supports_n: bool = False
#     supports_stop: bool = False

#     can_assistant_first: bool = False
#     can_assistant_last: bool = False
#     can_assistant_only: bool = False

#     can_system_last: bool = False

#     can_empty_content: bool = False
#     can_repeating_roles: bool = False
#     can_system_multiple: bool = False
#     can_min_p: bool = False
#     can_top_k: bool = False


# MODELS: list[Model] = [
#     PerplexityModel(
#         name="sonar",
#         cost=Cost(prompt=1.00e-6, completion=1.00e-6, request=0.005),
#         context_length=127000,
#         vendor_name=Vendor.perplexity.value,
#     ),
#     PerplexityModel(
#         name="sonar-pro",
#         cost=Cost(prompt=3.00e-6, completion=15.00e-6, request=0.005),
#         context_length=127000,
#         completion_length=8000,
#         vendor_name=Vendor.perplexity.value,
#     ),
#     PerplexityModel(
#         name="sonar-reasoning",
#         cost=Cost(prompt=1.00e-6, completion=5.00e-6, request=0.005),
#         context_length=127000,
#         vendor_name=Vendor.perplexity.value,
#     ),
# ]


# class PerplexitySDKChatProviderAdapter(OpenAISDKChatAdapter):
#     @staticmethod
#     def get_supported_models() -> list[Model]:
#         return MODELS

#     @staticmethod
#     def get_api_key_name() -> str:
#         return "PERPLEXITY_API_KEY"

#     def get_base_sdk_url(self) -> str:
#         return "https://api.perplexity.ai"



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import Cost, Model, Provider, Vendor


class XAIModel(Model):
    provider_name: str = Provider.xai.value
    vendor_name: str = Vendor.grok.value


MODELS: list[Model] = [
    XAIModel(
        name="grok-2",
        cost=Cost(prompt=2.00e-6, completion=10.00e-6),
        context_length=131072,
    ),
    # TODO: Add vision pricing
    XAIModel(
        name="grok-2-vision",
        cost=Cost(prompt=2.00e-6, completion=10.00e-6),
        context_length=32768,
    ),
]


class XAISDKChatProviderAdapter(OpenAISDKChatAdapter):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "XAI_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.x.ai/v1"



from os import PathLike
from typing import IO, Mapping, Optional, Tuple, TypedDict, Union
import httpx
from llm_adapters.types.openai import Body, Headers, Query, NotGiven, FileObject

FileContent = Union[IO[bytes], bytes, PathLike[str]]

FileTypes = Union[
    # file (or bytes)
    FileContent,
    # (filename, file (or bytes))
    Tuple[Optional[str], FileContent],
    # (filename, file (or bytes), content_type)
    Tuple[Optional[str], FileContent, Optional[str]],
    # (filename, file (or bytes), content_type, headers)
    Tuple[Optional[str], FileContent, Optional[str], Mapping[str, str]],
]


class ExtraKwargs(TypedDict, total=False):
    extra_headers: Headers | None
    extra_query: Query | None
    extra_body: Body | None
    timeout: float | httpx.Timeout | None | NotGiven


__all__ = ["FileObject", "ExtraKwargs", "FileTypes"]



import base64
import os
from typing import Any

import httpx

from llm_adapters.types import Cost

YUAN_TO_USD = 0.13737075

EMPTY_CONTENT = '""'


def delete_none_values(dictionary: dict[str, Any]) -> dict[str, Any]:
    if isinstance(dictionary, list):
        return [delete_none_values(e) for e in dictionary]

    if isinstance(dictionary, dict):
        for k, v in list(dictionary.items()):
            if v is None:
                del dictionary[k]
            else:
                dictionary[k] = delete_none_values(v)
    return dictionary


def process_image_url_anthropic(image_url: str) -> dict[str, Any]:
    httpx_client = httpx.Client()

    if image_url.startswith("data:"):
        media_type, _, base64_data = image_url.partition(";base64,")
        media_type = media_type.split(":")[1]
        return {
            "type": "image",
            "source": {
                "type": "base64",
                "media_type": media_type,
                "data": base64_data,
            },
        }
    else:
        image_data = base64.b64encode(httpx_client.get(image_url).content).decode(
            "utf-8"
        )

        _, extension = os.path.splitext(image_url)
        extension = extension.lstrip(".").lower()
        if extension == "jpg":
            extension = "jpeg"
        media_type = f"image/{extension}"

        return {
            "type": "image",
            "source": {
                "type": "base64",
                "media_type": media_type,
                "data": image_data,
            },
        }


def get_dynamic_cost(model_name: str, token_count: int) -> Cost:
    if model_name == "gemini-1.0-pro":
        return Cost(prompt=0.5e-6, completion=1.5e-6)
    elif model_name == "gemini-1.5-pro":
        if token_count <= 128000:
            return Cost(prompt=3.5e-6, completion=10.5e-6)
        else:
            return Cost(prompt=7.0e-6, completion=21.0e-6)
    elif model_name == "gemini-1.5-flash":
        if token_count <= 128000:
            return Cost(prompt=0.075e-6, completion=0.30e-6)
        else:
            return Cost(prompt=0.15e-6, completion=0.60e-6)
    else:
        raise ValueError(f"Unknown model: {model_name}")


class stream_generator_auto_close:
    _agen: Any

    def __init__(self, agen: Any) -> None:
        self._agen = agen

    async def __aenter__(self) -> Any:
        return self._agen

    async def __aexit__(self, *args: Any) -> None:
        if getattr(self._agen, "close", False):
            await self._agen.close()



# TODO: Remove this once openai types are updated
# mypy: ignore-errors

from typing import Mapping
from openai import NOT_GIVEN, NotGiven
from openai.types.completion import Completion
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)
from openai.types.chat import ChatCompletionMessageParam
from openai.types.chat.completion_create_params import (
    FunctionCall,
    Function as FunctionCreate,
    ResponseFormat,
)

from openai.types.chat.chat_completion_message_tool_call import (
    Function as FunctionToolCall,
)
from openai.types.chat.chat_completion_modality import ChatCompletionModality
from openai.types.chat.chat_completion_audio_param import ChatCompletionAudioParam
from openai.types.chat.chat_completion_prediction_content_param import (
    ChatCompletionPredictionContentParam,
)
from openai.types.chat.chat_completion_reasoning_effort import (
    ChatCompletionReasoningEffort,
)
from openai.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion_chunk import Choice as ChoiceChunk, ChoiceDelta
from openai.types.chat.chat_completion import (
    Choice,
    ChatCompletionMessage,
    CompletionUsage,
)
from openai.types.chat.chat_completion_user_message_param import (
    ChatCompletionUserMessageParam,
)
from openai.resources.files import FileObject
from openai.types.file_purpose import FilePurpose
from openai.types.shared_params.metadata import Metadata
from openai.resources.batches import Batch
from openai.pagination import SyncCursorPage, AsyncCursorPage
from openai._legacy_response import HttpxBinaryResponseContent
from openai._base_client import AsyncPaginator

Query = Mapping[str, object]
Body = object
Headers = Mapping[str, str]

ChatModel = str

__all__ = [
    "NotGiven",
    "NOT_GIVEN",
    "Completion",
    "ChatCompletion",
    "ChatCompletionChunk",
    "ChatCompletionMessageParam",
    "Query",
    "Body",
    "Headers",
    "ChatModel",
    "ChatCompletionAudioParam",
    "FunctionCall",
    "FunctionCreate",
    "FunctionToolCall",
    "ChatCompletionModality",
    "ChatCompletionPredictionContentParam",
    "ResponseFormat",
    "ChatCompletionReasoningEffort",
    "ChatCompletionStreamOptionsParam",
    "ChatCompletionToolChoiceOptionParam",
    "ChatCompletionToolParam",
    "ChatCompletionMessageToolCall",
    "Choice",
    "ChatCompletionMessage",
    "CompletionUsage",
    "ChatCompletionUserMessageParam",
    "ChoiceChunk",
    "ChoiceDelta",
    "FileObject",
    "FilePurpose",
    "Metadata",
    "Batch",
    "HttpxBinaryResponseContent",
    "SyncCursorPage",
    "AsyncCursorPage",
    "AsyncPaginator",
]



class AdapterException(Exception):
    pass


class AdapterModelNotFoundException(AdapterException):
    pass


class AdapterRateLimitException(AdapterException):
    pass


_all_ = [
    "AdapterException",
    "AdapterModelNotFoundException",
    "AdapterRateLimitException",
]



from typing import Any, Callable

from httpx import AsyncClient, Client, Limits, Timeout
from openai import AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.completion import Completion

from llm_adapters.abstract_adapters.sdk_chat_adapter import SDKChatAdapter
from llm_adapters.constants import (
    HTTP_CONNECT_TIMEOUT,
    HTTP_TIMEOUT,
    MAX_CONNECTIONS_PER_PROCESS,
    MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS,
)
from llm_adapters.types import (
    AdapterChatCompletion,
    AdapterChatCompletionChunk,
    AdapterCompletion,
    AdapterCompletionChunk,
)

CACHED_PROMPT_TOKEN_DISCOUNT = 0.5


class OpenAISDKChatAdapter(SDKChatAdapter[OpenAI, AsyncOpenAI]):
    def _call_completion_sync(self) -> Callable[..., Any]:
        return self._client_sync.completions.create

    def _call_completion_async(self) -> Callable[..., Any]:
        return self._client_async.completions.create

    def _call_sync(self) -> Callable[..., Any]:
        return self._client_sync.chat.completions.create

    def _call_async(self) -> Callable[..., Any]:
        return self._client_async.chat.completions.create

    def _create_client_sync(self, base_url: str, api_key: str) -> OpenAI:
        return OpenAI(
            base_url=base_url,
            api_key=api_key,
            max_retries=0,
            http_client=Client(
                limits=Limits(
                    max_connections=MAX_CONNECTIONS_PER_PROCESS,
                    max_keepalive_connections=MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS,
                ),
                timeout=Timeout(timeout=HTTP_TIMEOUT, connect=HTTP_CONNECT_TIMEOUT),
            ),
        )

    def _create_client_async(self, base_url: str, api_key: str) -> AsyncOpenAI:
        return AsyncOpenAI(
            base_url=base_url,
            api_key=api_key,
            max_retries=0,
            http_client=AsyncClient(
                limits=Limits(
                    max_connections=MAX_CONNECTIONS_PER_PROCESS,
                    max_keepalive_connections=MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS,
                ),
                timeout=Timeout(timeout=HTTP_TIMEOUT, connect=HTTP_CONNECT_TIMEOUT),
            ),
        )

    def _extract_response(
        self,
        request: Any,
        response: ChatCompletion,
    ) -> AdapterChatCompletion:
        prompt_tokens = response.usage.prompt_tokens if response.usage else 0
        completion_tokens = response.usage.completion_tokens if response.usage else 0
        reasoning_tokens = (
            response.usage.completion_tokens_details.reasoning_tokens
            if response.usage
            and response.usage.completion_tokens_details
            and response.usage.completion_tokens_details.reasoning_tokens
            else 0
        )

        cost = (
            self.get_model().cost.prompt * prompt_tokens
            + self.get_model().cost.completion * completion_tokens
            + reasoning_tokens * completion_tokens
            + self.get_model().cost.request
        )

        return AdapterChatCompletion.model_construct(**response.model_dump(), cost=cost)

    def _extract_stream_response(
        self, request: Any, response: ChatCompletionChunk, state: dict[str, Any]
    ) -> AdapterChatCompletionChunk:
        return AdapterChatCompletionChunk.model_construct(
            **response.model_dump(),
        )

    def _extract_completion_response(
        self,
        request: Any,
        response: Completion,
    ) -> AdapterCompletion:
        prompt_tokens = response.usage.prompt_tokens if response.usage else 0
        cached_prompt_tokens = (
            response.usage.prompt_tokens_details.cached_tokens
            if response.usage
            and response.usage.prompt_tokens_details
            and response.usage.prompt_tokens_details.cached_tokens
            else 0
        )
        completion_tokens = response.usage.completion_tokens if response.usage else 0
        reasoning_tokens = (
            response.usage.completion_tokens_details.reasoning_tokens
            if response.usage
            and response.usage.completion_tokens_details
            and response.usage.completion_tokens_details.reasoning_tokens
            else 0
        )

        uncached_prompt_tokens = prompt_tokens - cached_prompt_tokens

        cost = (
            uncached_prompt_tokens * self.get_model().cost.prompt
            + cached_prompt_tokens
            * self.get_model().cost.prompt
            * (1 - CACHED_PROMPT_TOKEN_DISCOUNT)
            + completion_tokens * self.get_model().cost.completion
            + reasoning_tokens * self.get_model().cost.completion
            + self.get_model().cost.request
        )

        return AdapterCompletion.model_construct(
            **response.model_dump(),
            cost=cost,
        )

    def _extract_completion_stream_response(
        self, request: Any, response: Completion, state: dict[str, Any]
    ) -> AdapterCompletionChunk:
        return AdapterCompletionChunk.model_construct(
            **response.model_dump(),
        )



from typing import (
    Any,
    AsyncGenerator,
    Generator,
    Optional,
    Union,
    Dict,
    Iterable,
    List,
    Literal,
)

from llm_adapters.types import (
    Completion,
    ChatCompletionChunk,
    ChatCompletion,
    Cost,
    ChatCompletionMessageToolCall,
    ChatCompletionMessageParam,
    ContentType,
)

from pydantic import BaseModel, ConfigDict
from enum import Enum
from openai.types.chat.chat_completion_message import FunctionCall

from llm_adapters.types.model import ConversationRole


class Turn(BaseModel, use_enum_values=True):
    role: Union[ConversationRole]
    content: str


class AdapterChatCompletion(ChatCompletion):
    cost: float

    # V7
    response: Optional[Turn] = None
    # V7
    token_counts: Optional[Cost] = None


class AdapterChatCompletionChunk(ChatCompletionChunk):
    pass


# V8
# AdapterStreamSyncChatCompletion = Generator[AdapterChatCompletionChunk, Any, None]
# V7
class AdapterStreamSyncChatCompletion:
    def __init__(
        self, response: Generator[AdapterChatCompletionChunk, Any, None]
    ) -> None:
        self.response = response

    def __iter__(self) -> "AdapterStreamSyncChatCompletion":
        return self

    def __next__(self) -> AdapterChatCompletionChunk:
        return next(self.response)

    def close(self) -> None:
        self.response.close()


# V8
# AdapterStreamAsyncChatCompletion = AsyncGenerator[AdapterChatCompletionChunk, Any]
# V7
class AdapterStreamAsyncChatCompletion:
    def __init__(
        self, response: AsyncGenerator[AdapterChatCompletionChunk, Any]
    ) -> None:
        self.response = response

    def __aiter__(self) -> "AdapterStreamAsyncChatCompletion":
        return self

    async def __anext__(self) -> AdapterChatCompletionChunk:
        return await self.response.__anext__()

    async def aclose(self) -> None:
        await self.response.aclose()


# V8
# AdapterStreamChatCompletion = Union[
#     AdapterStreamSyncChatCompletion,
#     AdapterStreamAsyncChatCompletion,
# ]
# V7
class AdapterStreamChatCompletion(BaseModel):
    response: Union[
        AdapterStreamSyncChatCompletion,
        AdapterStreamAsyncChatCompletion,
    ]

    model_config = ConfigDict(arbitrary_types_allowed=True)


class AdapterCompletion(Completion):
    cost: float


class AdapterCompletionChunk(Completion):
    pass


class AdapterStreamCompletion(BaseModel):
    response: Union[
        Generator[AdapterCompletionChunk, Any, None],
        AsyncGenerator[AdapterCompletionChunk, Any],
    ]

    model_config = ConfigDict(arbitrary_types_allowed=True)


class AdapterStreamSyncCompletion(AdapterStreamCompletion):
    response: Generator[AdapterCompletionChunk, Any, None]


class AdapterStreamAsyncCompletion(AdapterStreamCompletion):
    response: AsyncGenerator[AdapterCompletionChunk, Any]


# Deprecated


class FunctionOutputTurn(BaseModel, use_enum_values=True):
    role: Literal[ConversationRole.function] = ConversationRole.function
    content: Optional[str] = None
    name: str


class ToolOutputTurn(BaseModel, use_enum_values=True):
    role: Literal[ConversationRole.tool] = ConversationRole.tool
    content: Optional[str] = None
    tool_call_id: str


class FunctionCallTurn(BaseModel, use_enum_values=True):
    role: Literal[ConversationRole.assistant] = ConversationRole.assistant
    function_call: FunctionCall
    content: None = None


class ToolsCallTurn(BaseModel, use_enum_values=True):
    role: Literal[ConversationRole.assistant] = ConversationRole.assistant
    tool_calls: list[ChatCompletionMessageToolCall]
    content: None = None


class ImageDetailsType(str, Enum):
    high = "high"
    low = "low"
    auto = "auto"


class VisionImageDetails(BaseModel, use_enum_values=True):
    url: str
    details: Optional[ImageDetailsType] = ImageDetailsType.auto


class TextContentEntry(BaseModel, use_enum_values=True):
    type: Literal[ContentType.text] = ContentType.text
    text: str


class ImageContentEntry(BaseModel, use_enum_values=True):
    type: Literal[ContentType.image_url] = ContentType.image_url
    image_url: VisionImageDetails


class ContentTurn(BaseModel, use_enum_values=True, validate_assignment=True):
    role: str = ConversationRole.user
    content: List[Union[ImageContentEntry, TextContentEntry]]


TurnType = Union[
    Turn,
    FunctionOutputTurn,
    ToolOutputTurn,
    ToolsCallTurn,
    ContentTurn,
    FunctionCallTurn,
]


class Conversation(BaseModel):
    turns: List[TurnType]

    def __init__(
        self,
        turns: Union[
            "Conversation",
            List[TurnType],
            Dict[str, List[TurnType]],
        ],
    ):
        if isinstance(turns, Conversation):
            turns = turns.turns
        elif isinstance(turns, dict) and "turns" in turns:
            turns = turns["turns"]
        super().__init__(turns=turns)

    def __getitem__(self, index: int) -> TurnType:
        return self.turns[index]

    def __setitem__(self, index: int, value: TurnType) -> None:
        if not isinstance(value, Turn):
            raise ValueError("Value must be an instance of Turn")
        self.turns[index] = value

    def __len__(self) -> int:
        return len(self.turns)

    def __iter__(self) -> Any:
        return iter(self.turns)

    def is_last_turn_vision_query(self) -> bool:
        if len(self.turns):
            contentTurn = self.turns[len(self.turns) - 1]
        else:
            return False

        if isinstance(contentTurn, ContentTurn):
            return any(
                content.type == ContentType.image_url for content in contentTurn.content
            )

        return False

    def convert_to_prompt(self) -> "Prompt":
        return Prompt("".join([f"{turn.role}: {turn.content}" for turn in self.turns]))

    def convert_to_openai_format(self) -> Iterable[ChatCompletionMessageParam]:
        return self.model_dump()["turns"]  # type: ignore


# Deprecated
class Prompt(str):
    def convert_to_conversation(self) -> Conversation:
        return Conversation(turns=[Turn(role=ConversationRole.user, content=self)])


# Deprecated, Use AdapterChatCompletion
AdapterResponse = AdapterChatCompletion

# Deprecated, Use AdapterStreamChatCompletion
AdapterStreamResponse = AdapterStreamChatCompletion


__all__ = [
    "AdapterChatCompletion",
    "AdapterStreamChatCompletion",
    "AdapterStreamSyncChatCompletion",
    "AdapterStreamAsyncChatCompletion",
    "AdapterChatCompletionChunk",
    "AdapterCompletion",
    "AdapterCompletionChunk",
    "AdapterStreamCompletion",
    "AdapterStreamSyncCompletion",
    "AdapterStreamAsyncCompletion",
    # Deprecated
    "Turn",
    "FunctionOutputTurn",
    "ToolOutputTurn",
    "ToolsCallTurn",
    "ImageDetailsType",
    "VisionImageDetails",
    "TextContentEntry",
    "ImageContentEntry",
    "ContentTurn",
    "TurnType",
    "Conversation",
    "Prompt",
    "AdapterResponse",
    "AdapterStreamResponse",
]



# from typing import Any, Dict, Literal, Optional
# from vertexai.generative_models import GenerativeModel
# from google.cloud.aiplatform_v1.services.prediction_service import (
#     PredictionServiceClient,
#     PredictionServiceAsyncClient,
# )
# from google.api_core.client_options import ClientOptions
# from openai import NOT_GIVEN, NotGiven
# from openai.types.chat import ChatCompletionMessage
# from openai.types.chat.chat_completion import Choice

# from llm_adapters.abstract_adapters.api_key_adapter_mixin import ApiKeyAdapterMixin
# from llm_adapters.abstract_adapters.sdk_chat_adapter import SDKChatAdapter
# from llm_adapters.types import (
#     AdapterChatCompletion,
#     ContentTurn,
#     Conversation,
#     ConversationRole,
#     Cost,
#     Model,
#     ModelProperties,
#     Turn,
# )
# from llm_adapters.utils.general_utils import get_dynamic_cost

# PROVIDER_NAME = "gemini"
# API_KEY_NAME = "GEMINI_API_KEY"
# BASE_PROPERTIES = ModelProperties(gdpr_compliant=True)


# class GeminiModel(Model):

#     vendor_name: str = PROVIDER_NAME
#     provider_name: str = PROVIDER_NAME
#     properties: ModelProperties = BASE_PROPERTIES

#     supports_repeating_roles: bool = True
#     supports_system: bool = True
#     supports_multiple_system: bool = True
#     supports_empty_content: bool = True
#     supports_tool_choice_required: bool = True
#     supports_last_assistant: bool = True
#     supports_first_assistant: bool = True


# MODELS: list[Model] = [
#     GeminiModel(
#         name="gemini-1.0-pro",
#         cost=Cost(prompt=0.5e-6, completion=1.5e-6),
#         context_length=30720,
#         completion_length=2048,
#     ),
#     GeminiModel(
#         name="gemini-1.5-pro",
#         cost=Cost(prompt=3.5e-6, completion=10.5e-6),
#         context_length=128000,
#         completion_length=8192,
#     ),
#     GeminiModel(
#         name="gemini-1.5-flash",
#         cost=Cost(prompt=0.35e-6, completion=0.70e-6),
#         context_length=128000,
#         completion_length=8192,
#     ),
# ]


# # TODO: max_tokens doesnt work
# class GeminiSDKChatProviderAdapter(
#     ProviderAdapterMixin,
#     ApiKeyAdapterMixin,
#     SDKChatAdapter,
# ):
#     @staticmethod
#     def get_supported_models() -> list[Model]:
#         return MODELS

#     def get_base_sdk_url(self) -> str:
#         return ""

#     @staticmethod
#     def get_api_key_name() -> str:
#         return API_KEY_NAME

#     _sync_client: PredictionServiceClient
#     _async_client: PredictionServiceAsyncClient

#     def __init__(
#         self,
#     ) -> None:
#         super().__init__()
#         self.set_api_key(self.get_api_key())
#         self.model: GenerativeModel | None = None

#     def get_model_name(self) -> str:
#         if self._current_model is None:
#             raise ValueError("Model not set")
#         return self._current_model.name

#     def get_async_client(self):
#         return self._async_client

#     def get_sync_client(self):
#         return self._sync_client

#     def adjust_temperature(self, temperature: float) -> float:
#         return temperature / 2

#     def set_api_key(self, api_key: str) -> None:
#         super().set_api_key(api_key)

#         self._sync_client = PredictionServiceClient(
#             client_options=ClientOptions(api_key=api_key), transport="rest"
#         )
#         self._async_client = PredictionServiceAsyncClient(
#             client_options=ClientOptions(api_key=api_key)
#         )

#     def extract_response(self, request: Any, response: Any) -> AdapterChatCompletion:
#         choices = [
#             Choice(
#                 message=ChatCompletionMessage(
#                     role=ConversationRole.assistant.value,
#                     content=response.text,
#                 ),
#                 finish_reason="stop",
#                 index=0,
#             )
#         ]

#         # Optimize token count calculation, use async for async and parallelize
#         assert self.model
#         prompt_tokens = self.model.count_tokens(
#             [_map_turn_content_to_str(turn) for turn in request.turns]
#         ).total_tokens
#         completion_tokens = self.model.count_tokens(response.text).total_tokens

#         dynamic_cost = get_dynamic_cost(self.get_model_name(), prompt_tokens)
#         cost = (
#             dynamic_cost.prompt * prompt_tokens
#             + dynamic_cost.completion * completion_tokens
#             + self.get_model().cost.request
#         )

#         return AdapterChatCompletion(
#             response=Turn(
#                 role=ConversationRole.assistant,
#                 content=choices[0]["message"]["content"],  # type: ignore
#             ),  # TODO: Refactor response
#             choices=choices,
#             cost=cost,
#             token_counts=Cost(
#                 prompt=prompt_tokens,
#                 completion=completion_tokens,
#             ),
#         )

#     async def extract_response_async(
#         self, request: Any, response: Any
#     ) -> AdapterChatCompletion:
#         model = GenerativeModel(model_name=self.get_model_name())
#         model._async_client = self.get_async_client()

#         choices = [
#             Choice(
#                 message=ChatCompletionMessage(
#                     role=ConversationRole.assistant.value,
#                     content=response.text,
#                 ),
#                 finish_reason="stop",
#                 index=0,
#             )
#         ]

#         prompt_tokens = await model.count_tokens_async(
#             [turn.content for turn in request.turns]
#         )
#         completion_tokens = await model.count_tokens_async(response.text)

#         dynamic_cost = get_dynamic_cost(self.get_model_name(), prompt_tokens)
#         cost = (
#             dynamic_cost.prompt * prompt_tokens
#             + dynamic_cost.completion * completion_tokens
#             + self.get_model().cost.request
#         )

#         return AdapterChatCompletion(
#             response=Turn(
#                 role=ConversationRole.assistant,
#                 content=choices[0]["message"]["content"],  # type: ignore
#             ),
#             choices=choices,
#             cost=cost,
#             token_counts=Cost(
#                 prompt=prompt_tokens.total_tokens,
#                 completion=completion_tokens.total_tokens,
#             ),
#         )

#     def extract_stream_response(self, request, response):
#         return response

#     def get_params(
#         self,
#         llm_input: Conversation,
#         **kwargs: Any,
#     ) -> Dict[str, Any]:
#         params = super().get_params(llm_input, **kwargs)

#         last_message = params["messages"].pop()
#         last_content = last_message["content"] or " "

#         transformed_messages = []
#         for message in params["messages"]:
#             transformed_message = {
#                 "parts": [_map_content_to_str(message["content"], message["role"])],
#                 "role": _map_role(message["role"]),
#             }
#             transformed_messages.append(transformed_message)

#         if "max_tokens" in params:
#             params["max_output_tokens"] = params["max_tokens"]
#             del params["max_tokens"]

#         del params["messages"]

#         return {
#             "config": params,
#             "history": transformed_messages,
#             "prompt": _map_content_to_str(last_content, last_message["role"]),
#         }

#     async def execute_async(
#         self,
#         llm_input: Conversation,
#         stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
#         **kwargs: Any,
#     ):
#         params = self.get_params(llm_input, **kwargs)

#         model = GenerativeModel(
#             model_name=self.get_model_name(),
#             generation_config=params["config"],
#         )
#         model._prediction_async_client_value = self.get_async_client()

#         convo = model.start_chat(history=params["history"])

#         result = await convo.send_message_async(params["prompt"])

#         return await self.extract_response_async(request=llm_input, response=result)

#     def execute_sync(
#         self,
#         llm_input: Conversation,
#         stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
#         **kwargs: Any,
#     ):
#         params = self.get_params(llm_input, **kwargs)

#         self.model = GenerativeModel(
#             model_name=self.get_model_name(),
#             generation_config=params.get("config") or None,
#         )
#         self.model._prediction_client_value = self.get_sync_client()

#         chat = self.model.start_chat(history=params["history"])
#         result = chat.send_message(params["prompt"])
#         return self.extract_response(request=llm_input, response=result)


# def _map_content_to_str(
#     content: str | list[dict[str, Any]], role: str | ConversationRole
# ) -> str:
#     if not content or content in [" ", "\n", "\t"]:
#         return "."

#     # Join content if it is a list.
#     if isinstance(content, list):
#         return " ".join(content.get("text", "") for content in content)

#     if role != "system":
#         return content

#     # Simulate *system message*.
#     return f"*{content}*"


# def _map_role(role: str | ConversationRole) -> str:
#     match role:
#         case ConversationRole.user | ConversationRole.system | "user" | "system":
#             return "user"
#         case ConversationRole.assistant | "assistant":
#             return "model"
#         case _:
#             return "user"


# def _map_turn_content_to_str(turn: ContentTurn | Turn) -> str:
#     match turn:
#         case ContentTurn():
#             return " ".join(content.text for content in turn.content)  # type: ignore[union-attr]
#         case Turn():
#             return turn.content



from typing import Any, Iterable, Literal, Optional, Union, Unpack, cast, overload
from openai import OpenAI as OOpenAI
from openai import AsyncOpenAI as AsyncOOpenAI
from llm_adapters.adapter_factory import AdapterFactory
from llm_adapters.types import (
    ChatCompletionCreateKwargs,
    CompletionCreateKwargs,
    NotGiven,
    NOT_GIVEN,
    AdapterChatCompletion,
    AdapterStreamSyncChatCompletion,
    ChatModel,
    ChatCompletionMessageParam,
    AdapterCompletion,
    AdapterStreamSyncCompletion,
    ChatCompletionCreateArgs,
    CompletionCreateArgs,
    AdapterStreamAsyncChatCompletion,
    AdapterStreamAsyncCompletion,
    FileObject,
    FilePurpose,
    Metadata,
    Batch,
    HttpxBinaryResponseContent,
    SyncCursorPage,
    AsyncCursorPage,
    AsyncPaginator,
)
from llm_adapters.types.files import ExtraKwargs, FileTypes


class Completions:
    @overload
    def create(
        self,
        *,
        model: Union[str, ChatModel],
        prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None],
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateKwargs],
    ) -> AdapterCompletion: ...

    @overload
    def create(
        self,
        *,
        model: Union[str, ChatModel],
        prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None],
        stream: Literal[True],
        **kwargs: Unpack[CompletionCreateKwargs],
    ) -> AdapterStreamSyncCompletion: ...

    def create(
        self,
        *,
        model: Union[str, ChatModel],
        prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None],
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateKwargs],
    ) -> AdapterCompletion | AdapterStreamSyncCompletion:
        return AdapterFactory.get_adapter_by_path(model).execute_completion_sync(
            stream=stream,
            **cast(CompletionCreateArgs, {"model": model, "prompt": prompt, **kwargs}),
        )


class AsyncCompletions:
    @overload
    async def create(
        self,
        *,
        model: Union[str, ChatModel],
        prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None],
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateKwargs],
    ) -> AdapterCompletion: ...

    @overload
    async def create(
        self,
        *,
        model: Union[str, ChatModel],
        prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None],
        stream: Literal[True],
        **kwargs: Unpack[CompletionCreateKwargs],
    ) -> AdapterStreamAsyncCompletion: ...

    async def create(
        self,
        *,
        model: Union[str, ChatModel],
        prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None],
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateKwargs],
    ) -> AdapterCompletion | AdapterStreamAsyncCompletion:
        return await AdapterFactory.get_adapter_by_path(model).execute_completion_async(
            stream=stream,
            **cast(CompletionCreateArgs, {"model": model, "prompt": prompt, **kwargs}),
        )


class ChatCompletions:
    @overload
    def create(
        self,
        *,
        model: Union[str, ChatModel],
        messages: Iterable[ChatCompletionMessageParam],
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateKwargs],
    ) -> AdapterChatCompletion: ...

    @overload
    def create(
        self,
        *,
        model: Union[str, ChatModel],
        messages: Iterable[ChatCompletionMessageParam],
        stream: Literal[True],
        **kwargs: Unpack[ChatCompletionCreateKwargs],
    ) -> AdapterStreamSyncChatCompletion: ...

    def create(
        self,
        *,
        model: Union[str, ChatModel],
        messages: Iterable[ChatCompletionMessageParam],
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateKwargs],
    ) -> AdapterChatCompletion | AdapterStreamSyncChatCompletion:
        return AdapterFactory.get_adapter_by_path(model).execute_chat_completion_sync(
            stream=stream,
            **cast(
                ChatCompletionCreateArgs,
                {"model": model, "messages": messages, **kwargs},
            ),
        )


class AsyncChatCompletions:
    @overload
    async def create(
        self,
        *,
        model: Union[str, ChatModel],
        messages: Iterable[ChatCompletionMessageParam],
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateKwargs],
    ) -> AdapterChatCompletion: ...

    @overload
    async def create(
        self,
        *,
        model: Union[str, ChatModel],
        messages: Iterable[ChatCompletionMessageParam],
        stream: Literal[True],
        **kwargs: Unpack[ChatCompletionCreateKwargs],
    ) -> AdapterStreamAsyncChatCompletion: ...

    async def create(
        self,
        *,
        model: Union[str, ChatModel],
        messages: Iterable[ChatCompletionMessageParam],
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateKwargs],
    ) -> AdapterChatCompletion | AdapterStreamAsyncChatCompletion:
        return await AdapterFactory.get_adapter_by_path(
            model
        ).execute_chat_completion_async(
            stream=stream,
            **cast(
                ChatCompletionCreateArgs,
                {"model": model, "messages": messages, **kwargs},
            ),
        )


class Batches:
    def create(
        self,
        *,
        completion_window: Literal["24h"],
        endpoint: Literal["/v1/chat/completions", "/v1/embeddings", "/v1/completions"],
        input_file_id: str,
        metadata: Optional[Metadata] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ExtraKwargs],
    ) -> Batch:
        return OOpenAI().batches.create(
            completion_window=completion_window,
            endpoint=endpoint,
            input_file_id=input_file_id,
            metadata=metadata,
            **kwargs,
        )

    def retrieve(self, batch_id: str, **kwargs: Unpack[ExtraKwargs]) -> Batch:
        return OOpenAI().batches.retrieve(batch_id, **kwargs)

    def cancel(self, batch_id: str, **kwargs: Unpack[ExtraKwargs]) -> Batch:
        return OOpenAI().batches.cancel(batch_id, **kwargs)

    def list(
        self,
        *,
        after: str | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ExtraKwargs],
    ) -> SyncCursorPage[Batch]:
        return OOpenAI().batches.list(after=after, limit=limit, **kwargs)


class AsyncBatches:
    async def create(
        self,
        *,
        completion_window: Literal["24h"],
        endpoint: Literal["/v1/chat/completions", "/v1/embeddings", "/v1/completions"],
        input_file_id: str,
        metadata: Optional[Metadata] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ExtraKwargs],
    ) -> Batch:
        return await AsyncOOpenAI().batches.create(
            completion_window=completion_window,
            endpoint=endpoint,
            input_file_id=input_file_id,
            metadata=metadata,
            **kwargs,
        )

    async def retrieve(self, batch_id: str, **kwargs: Unpack[ExtraKwargs]) -> Batch:
        return await AsyncOOpenAI().batches.retrieve(batch_id, **kwargs)

    async def cancel(self, batch_id: str, **kwargs: Unpack[ExtraKwargs]) -> Batch:
        return await AsyncOOpenAI().batches.cancel(batch_id, **kwargs)

    async def list(
        self,
        *,
        after: str | NotGiven = NOT_GIVEN,
        limit: int | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ExtraKwargs],
    ) -> AsyncPaginator[Batch, AsyncCursorPage[Batch]]:
        return AsyncOOpenAI().batches.list(after=after, limit=limit, **kwargs)


class Files:
    def create(
        self,
        *,
        file: FileTypes,
        purpose: FilePurpose,
        **kwargs: Unpack[ExtraKwargs],
    ) -> FileObject:
        return OOpenAI().files.create(file=file, purpose=purpose, **kwargs)

    def content(
        self, file_id: str, **kwargs: Unpack[ExtraKwargs]
    ) -> HttpxBinaryResponseContent:
        return OOpenAI().files.content(file_id, **kwargs)


class AsyncFiles:
    async def create(
        self,
        *,
        file: FileTypes,
        purpose: FilePurpose,
        **kwargs: Unpack[ExtraKwargs],
    ) -> FileObject:
        return await AsyncOOpenAI().files.create(file=file, purpose=purpose, **kwargs)

    async def content(
        self, file_id: str, **kwargs: Unpack[ExtraKwargs]
    ) -> HttpxBinaryResponseContent:
        return await AsyncOOpenAI().files.content(file_id, **kwargs)


class Chat:
    def __init__(self) -> None:
        self._chat_completions = ChatCompletions()

    @property
    def completions(self) -> ChatCompletions:
        return self._chat_completions


class AsyncChat:
    def __init__(self) -> None:
        self._async_chat_completions = AsyncChatCompletions()

    @property
    def completions(self) -> AsyncChatCompletions:
        return self._async_chat_completions


class OpenAI:
    def __init__(self, **kwargs: Any) -> None:
        self._chat = Chat()
        self._completions = Completions()
        self._files = Files()
        self._batches = Batches()

    @property
    def chat(self) -> Chat:
        return self._chat

    @property
    def completions(self) -> Completions:
        return self._completions

    @property
    def files(self) -> Files:
        return self._files

    @property
    def batches(self) -> Batches:
        return self._batches


class AsyncOpenAI:
    def __init__(self, **kwargs: Any) -> None:
        self._async_chat = AsyncChat()
        self._async_completions = AsyncCompletions()
        self._async_files = AsyncFiles()
        self._async_batches = AsyncBatches()

    @property
    def chat(self) -> AsyncChat:
        return self._async_chat

    @property
    def completions(self) -> AsyncCompletions:
        return self._async_completions

    @property
    def files(self) -> AsyncFiles:
        return self._async_files

    @property
    def batches(self) -> AsyncBatches:
        return self._async_batches


__all__ = ["OpenAI", "AsyncOpenAI"]



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import Cost, Model, Provider, Vendor


class LambdaLabsModel(Model):
    provider_name: str = Provider.lambdalabs.value

    supports_completion: bool = False

    supports_n: bool = False
    supports_json_output: bool = False
    supports_stop: bool = False
    supports_vision: bool = False
    supports_tools: bool = False

    can_min_p: bool = False
    can_top_k: bool = False

    def _get_api_path(self) -> str:
        return f"{self.name}"


MODELS: list[Model] = [
    LambdaLabsModel(
        name="llama3.1-405b-instruct-fp8",
        vendor_name=Vendor.meta_llama.value,
        cost=Cost(prompt=0.90e-6, completion=0.90e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="lfm-40b",
        vendor_name=Vendor.liquid.value,
        cost=Cost(prompt=0.15e-6, completion=0.15e-6),
        context_length=66000,
        supports_max_completion_tokens=False,
    ),
    LambdaLabsModel(
        name="llama3.1-8b-instruct",
        vendor_name=Vendor.meta_llama.value,
        cost=Cost(prompt=0.03e-6, completion=0.03e-6),
        context_length=131000,
    ),
    # LambdaLabsModel(
    #     name="llama3.2-11b-vision-instruct",
    #     vendor_name=Vendor.meta_llama.value,
    #     cost=Cost(
    #         prompt=0.0e-6, completion=0.0e-6
    #     ),  # pricing not provided in the table
    #     context_length=131000,
    # ),
    LambdaLabsModel(
        name="qwen25-coder-32b-instruct",
        vendor_name=Vendor.qwen.value,
        cost=Cost(prompt=0.09e-6, completion=0.09e-6),
        context_length=33000,
    ),
    LambdaLabsModel(
        name="hermes3-405b",
        vendor_name=Vendor.nous_hermes.value,
        cost=Cost(prompt=0.90e-6, completion=0.90e-6),
        context_length=131000,
        can_assistant_only=False,
    ),
    LambdaLabsModel(
        name="deepseek-llama3.3-70b",
        vendor_name=Vendor.deepseek.value,
        cost=Cost(prompt=0.20e-6, completion=0.20e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="llama3.1-70b-instruct-fp8",
        vendor_name=Vendor.meta_llama.value,
        cost=Cost(prompt=0.20e-6, completion=0.20e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="llama3.3-70b-instruct-fp8",
        vendor_name=Vendor.meta_llama.value,
        cost=Cost(prompt=0.20e-6, completion=0.20e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="llama3.2-3b-instruct",
        vendor_name=Vendor.meta_llama.value,
        cost=Cost(prompt=0.02e-6, completion=0.02e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="hermes3-8b",
        vendor_name=Vendor.nous_hermes.value,
        cost=Cost(prompt=0.03e-6, completion=0.03e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="llama3.1-nemotron-70b-instruct-fp8",
        vendor_name=Vendor.meta_llama.value,
        cost=Cost(prompt=0.20e-6, completion=0.20e-6),
        context_length=131000,
    ),
    LambdaLabsModel(
        name="hermes3-70b",
        vendor_name=Vendor.nous_hermes.value,
        cost=Cost(prompt=0.20e-6, completion=0.20e-6),
        context_length=131000,
        can_assistant_only=False,
    ),
]


class LambdaLabsSDKChatProviderAdapter(OpenAISDKChatAdapter):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "LAMBDALABS_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.lambdalabs.com/v1"



from typing import (
    Iterable,
    Literal,
    Optional,
    TypedDict,
    Union,
)

import httpx

from llm_adapters.types import (
    NotGiven,
    ChatCompletionMessageParam,
    ChatModel,
    ChatCompletionAudioParam,
    FunctionCall,
    FunctionCreate,
    ChatCompletionModality,
    ChatCompletionPredictionContentParam,
    ChatCompletionReasoningEffort,
    ResponseFormat,
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)

from llm_adapters.types.openai import Body, Headers, Query


class CompletionCreateKwargs(TypedDict, total=False):
    # model: Union[str]
    # prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None]
    best_of: Optional[int] | NotGiven
    echo: Optional[bool] | NotGiven
    frequency_penalty: Optional[float] | NotGiven
    logit_bias: Optional[dict[str, int]] | NotGiven
    logprobs: Optional[int] | NotGiven
    max_tokens: Optional[int] | NotGiven
    n: Optional[int] | NotGiven
    presence_penalty: Optional[float] | NotGiven
    seed: Optional[int] | NotGiven
    stop: Union[Optional[str], list[str], None] | NotGiven
    # stream: Optional[Literal[False]] | NotGiven
    stream_options: Optional[ChatCompletionStreamOptionsParam] | NotGiven
    suffix: Optional[str] | NotGiven
    temperature: Optional[float] | NotGiven
    top_p: Optional[float] | NotGiven
    user: str | NotGiven

    extra_headers: Headers | None
    extra_query: Query | None
    extra_body: Body | None
    timeout: float | httpx.Timeout | None | NotGiven


class CompletionCreateArgs(CompletionCreateKwargs, total=False):
    model: Union[str]
    prompt: Union[str, list[str], Iterable[int], Iterable[Iterable[int]], None]
    # stream: Optional[Literal[False]] | NotGiven


class ChatCompletionCreateKwargs(TypedDict, total=False):
    # messages: Iterable[ChatCompletionMessageParam]
    # model: Union[str, ChatModel]
    audio: Optional[ChatCompletionAudioParam] | None
    frequency_penalty: Optional[float] | None
    function_call: FunctionCall | None
    functions: Iterable[FunctionCreate] | None
    logit_bias: Optional[dict[str, int]] | None
    logprobs: Optional[bool] | None
    max_completion_tokens: Optional[int] | None
    max_tokens: Optional[int] | None
    metadata: Optional[dict[str, str]] | None
    modalities: Optional[list[ChatCompletionModality]] | None
    n: Optional[int] | None
    parallel_tool_calls: bool | None
    prediction: Optional[ChatCompletionPredictionContentParam] | None
    presence_penalty: Optional[float] | None
    reasoning_effort: ChatCompletionReasoningEffort | None
    response_format: ResponseFormat | None
    seed: Optional[int] | None
    service_tier: Optional[Literal["auto", "default"]] | None
    stop: Union[Optional[str], list[str]] | None
    store: Optional[bool] | None
    # stream: Optional[Literal[False]] | None
    stream_options: Optional[ChatCompletionStreamOptionsParam] | None
    temperature: Optional[float] | None
    tool_choice: ChatCompletionToolChoiceOptionParam | None
    tools: Iterable[ChatCompletionToolParam] | None
    top_logprobs: Optional[int] | None
    top_p: Optional[float] | None
    user: str | None

    # min_p: Optional[float] | None

    extra_headers: Headers | None
    extra_query: Query | None
    extra_body: Body | None
    timeout: float | httpx.Timeout | None


class ChatCompletionCreateArgs(ChatCompletionCreateKwargs):
    messages: Iterable[ChatCompletionMessageParam]
    model: Union[str, ChatModel]
    # stream: Optional[Literal[False]] | NotGiven


__all__ = [
    "ChatCompletionCreateKwargs",
    "CompletionCreateKwargs",
    "ChatCompletionCreateArgs",
    "CompletionCreateArgs",
]



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import Cost, Model, Provider, Vendor


class OpenAIModel(Model):
    provider_name: str = Provider.openai.value
    vendor_name: str = Vendor.openai.value

    supports_completion: bool = False

    can_min_p: bool = False
    can_top_k: bool = False


MODELS: list[Model] = [
    OpenAIModel(
        name="gpt-3.5-turbo-instruct",
        cost=Cost(prompt=1.50e-6, completion=2.00e-6),
        context_length=16385,
        completion_length=16385,
        supports_completion=True,
        supports_chat=False,
        supports_max_completion_tokens=False,
    ),
    OpenAIModel(
        name="gpt-4o",
        cost=Cost(prompt=2.50e-6, completion=10.00e-6),
        context_length=128000,
        completion_length=16384,
    ),
    OpenAIModel(
        name="gpt-4o-2024-11-20",
        cost=Cost(prompt=2.50e-6, completion=10.00e-6),
        context_length=128000,
        completion_length=16384,
    ),
    OpenAIModel(
        name="gpt-4o-2024-08-06",
        cost=Cost(prompt=2.50e-6, completion=10.00e-6),
        context_length=128000,
        completion_length=16384,
    ),
    OpenAIModel(
        name="gpt-4o-2024-11-20",
        cost=Cost(prompt=2.50e-6, completion=10.00e-6),
        context_length=128000,
        completion_length=16384,
    ),
    OpenAIModel(
        name="gpt-4o-mini",
        cost=Cost(prompt=0.15e-6, completion=0.60e-6),
        context_length=128000,
        completion_length=16385,
    ),
    OpenAIModel(
        name="gpt-4o-mini-2024-07-18",
        cost=Cost(prompt=0.15e-6, completion=0.60e-6),
        context_length=128000,
        completion_length=16385,
    ),
    OpenAIModel(
        name="o1-mini",
        cost=Cost(prompt=1.10e-6, completion=4.40e-6),
        context_length=128000,
        completion_length=65536,
        supports_json_output=False,
        supports_stop=False,
        supports_tools=False,
        can_temperature=False,
        can_system=False,
    ),
    OpenAIModel(
        name="o1-mini-2024-09-12",
        cost=Cost(prompt=1.10e-6, completion=4.40e-6),
        context_length=128000,
        completion_length=65536,
        supports_json_output=False,
        supports_stop=False,
        supports_tools=False,
        can_temperature=False,
        can_system=False,
    ),
    OpenAIModel(
        name="o1",
        cost=Cost(prompt=15.00e-6, completion=60.00e-6),
        context_length=200000,
        completion_length=100000,
        can_temperature=False,
    ),
    OpenAIModel(
        name="o1-2024-12-17",
        cost=Cost(prompt=15.00e-6, completion=60.00e-6),
        context_length=200000,
        completion_length=100000,
        can_temperature=False,
    ),
    OpenAIModel(
        name="o3-mini",
        cost=Cost(prompt=1.10e-6, completion=4.40e-6),
        context_length=200000,
        completion_length=100000,
        can_temperature=False,
    ),
    OpenAIModel(
        name="o3-mini-2025-01-31",
        cost=Cost(prompt=1.10e-6, completion=4.40e-6),
        context_length=200000,
        completion_length=100000,
        can_temperature=False,
    ),
    OpenAIModel(
        name="gpt-4.5-preview-2025-02-27",
        cost=Cost(prompt=75.00e-6, completion=150.00e-6),
        context_length=128000,
        completion_length=16384,
        supports_n=False,
        supports_stop=False,
        can_system_last=False,
        can_repeating_roles=False,
        can_empty_content=False,
        can_temperature=False,
    ),
    OpenAIModel(
        name="gpt-4.5-preview",
        cost=Cost(prompt=75.00e-6, completion=150.00e-6),
        context_length=128000,
        completion_length=16384,
        supports_n=False,
        supports_stop=False,
        can_system_last=False,
        can_repeating_roles=False,
        can_empty_content=False,
        can_temperature=False,
    ),
    # OpenAIModel(
    #     name="o1-pro",
    #     cost=Cost(prompt=150.00e-6, completion=600.00e-6),
    #     context_length=200000,
    #     completion_length=100000,
    #     is_response=True,
    # ),
    # OpenAIModel(
    #     name="o1-pro-2025-03-19",
    #     cost=Cost(prompt=150.00e-6, completion=600.00e-6),
    #     context_length=200000,
    #     completion_length=100000,
    #     is_response=True,
    # ),
]


class OpenAISDKChatProviderAdapter(OpenAISDKChatAdapter):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "OPENAI_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.openai.com/v1"



from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
from llm_adapters.types import Cost, Model, Provider, Vendor


class DeepSeekModel(Model):
    provider_name: str = Provider.deepseek.value
    vendor_name: str = Vendor.deepseek_ai.value

    supports_vision: bool = False
    supports_n: bool = False

    def _get_api_path(self) -> str:
        return f"{self.name}"


MODELS: list[Model] = [
    DeepSeekModel(
        name="deepseek-chat",
        cost=Cost(prompt=0.14e-6, completion=0.28e-6),
        context_length=64000,
        completion_length=8000,
        supports_completion=False,  # Available in beta mode
    ),
    DeepSeekModel(
        name="deepseek-reasoner",
        cost=Cost(prompt=0.55e-6, completion=2.19e-6),
        context_length=64000,
        supports_json_output=False,
        supports_completion=False,
        supports_max_tokens=False,
        supports_tools=False,
        can_assistant_first=False,
        can_assistant_last=False,
        can_assistant_only=False,
        can_repeating_roles=False,
        can_system_last=False,
        can_system_only=False,
        can_system_multiple=False,
    ),
]


class DeepSeekSDKChatProviderAdapter(OpenAISDKChatAdapter):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "DEEPSEEK_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.deepseek.com"



# from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
# from llm_adapters.types import Cost, Model, Provider, Vendor


# class MoescapeModel(Model):
#     provider_name: str = Provider.lepton.value
#     vendor_name: str = Vendor.moescape.value


# MODELS: list[Model] = [
#     MoescapeModel(
#         name="nephara",
#         api_name="llama3-8b-instruct",
#         cost=Cost(prompt=0, completion=0),
#         context_length=8192,
#     ),
#     MoescapeModel(
#         name="lunara",
#         api_name="lunara-12bv1",
#         cost=Cost(prompt=0, completion=0),
#         context_length=12288,
#         can_system_only=False,
#     ),
# ]


# class MoescapeSDKChatProviderAdapter(OpenAISDKChatAdapter):
#     @staticmethod
#     def get_supported_models() -> list[Model]:
#         return MODELS

#     @staticmethod
#     def get_api_key_name() -> str:
#         return "MOESCAPE_API_KEY"

#     def get_base_sdk_url(self) -> str:
#         return f"https://jtwtcue2-{self._current_model.api_name if self._current_model and self._current_model.api_name else ''}.tin.lepton.run/api/v1/"

#     def _set_current_model(self, model: Model) -> None:
#         super()._set_current_model(model)

#         self._setup_clients(self.get_api_key())



from abc import ABC, abstractmethod
from typing import Any, Iterable, Literal, Optional, Unpack, overload

from openai import NOT_GIVEN, NotGiven

from llm_adapters.types import (
    AdapterChatCompletion,
    AdapterCompletion,
    AdapterStreamAsyncChatCompletion,
    AdapterStreamAsyncCompletion,
    AdapterStreamSyncChatCompletion,
    AdapterStreamSyncCompletion,
    Model,
    ChatCompletionMessageParam,
    Conversation,
)
from llm_adapters.types.request import (
    ChatCompletionCreateArgs,
    CompletionCreateArgs,
)

# TODO: add Deepseek and pricing

# DeepSeek-V3 at 50% off
# DeepSeek-R1 at a massive 75% off


class Adapter(ABC):
    def __str__(self) -> str:
        return f"adapter-{self.get_model().get_path()}"

    def set_api_key(self, api_key: str) -> None:
        pass

    @abstractmethod
    def get_model(self) -> Model:
        pass

    @overload
    def execute_completion_sync(
        self,
        *,
        stream: Literal[True],
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterStreamSyncCompletion: ...
    @overload
    def execute_completion_sync(
        self,
        *,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion: ...
    @abstractmethod
    def execute_completion_sync(
        self,
        *,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion | AdapterStreamSyncCompletion: ...

    @overload
    async def execute_completion_async(
        self,
        *,
        stream: Literal[True],
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterStreamAsyncCompletion: ...
    @overload
    async def execute_completion_async(
        self,
        *,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion: ...
    @abstractmethod
    async def execute_completion_async(
        self,
        *,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[CompletionCreateArgs],
    ) -> AdapterCompletion | AdapterStreamAsyncCompletion: ...

    @overload
    def execute_chat_completion_sync(
        self,
        *,
        stream: Literal[True],
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterStreamSyncChatCompletion: ...
    @overload
    def execute_chat_completion_sync(
        self,
        *,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion: ...
    @abstractmethod
    def execute_chat_completion_sync(
        self,
        *,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion | AdapterStreamSyncChatCompletion: ...

    @overload
    async def execute_chat_completion_async(
        self,
        *,
        stream: Literal[True],
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterStreamAsyncChatCompletion: ...
    @overload
    async def execute_chat_completion_async(
        self,
        *,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion: ...
    @abstractmethod
    async def execute_chat_completion_async(
        self,
        *,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Unpack[ChatCompletionCreateArgs],
    ) -> AdapterChatCompletion | AdapterStreamAsyncChatCompletion: ...

    # V7
    @overload
    def execute_sync(
        self,
        messages: Iterable[ChatCompletionMessageParam] | Conversation,
        stream: Literal[True],
        **kwargs: Any,
    ) -> AdapterStreamSyncChatCompletion: ...
    @overload
    def execute_sync(
        self,
        messages: Iterable[ChatCompletionMessageParam] | Conversation,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> AdapterChatCompletion: ...
    def execute_sync(
        self,
        messages: Iterable[ChatCompletionMessageParam] | Conversation,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> AdapterChatCompletion | AdapterStreamSyncChatCompletion:
        if isinstance(messages, Conversation):
            openai_messages = messages.convert_to_openai_format()
        else:
            openai_messages = messages

        return self.execute_chat_completion_sync(
            messages=openai_messages, stream=stream, **kwargs
        )

    # V7
    @overload
    async def execute_async(
        self,
        messages: Iterable[ChatCompletionMessageParam] | Conversation,
        stream: Literal[True],
        **kwargs: Any,
    ) -> AdapterStreamAsyncChatCompletion: ...
    @overload
    async def execute_async(
        self,
        messages: Iterable[ChatCompletionMessageParam] | Conversation,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> AdapterChatCompletion: ...
    async def execute_async(
        self,
        messages: Iterable[ChatCompletionMessageParam] | Conversation,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> AdapterChatCompletion | AdapterStreamAsyncChatCompletion:
        if isinstance(messages, Conversation):
            openai_messages = messages.convert_to_openai_format()
        else:
            openai_messages = messages

        return await self.execute_chat_completion_async(
            messages=openai_messages, stream=stream, **kwargs
        )



# from typing import Any, Dict

# from llm_adapters.abstract_adapters.openai_sdk_chat_adapter import OpenAISDKChatAdapter
# from llm_adapters.types import (
#     Conversation,
#     ConversationRole,
#     Cost,
#     Model,
#     Provider,
#     Vendor,
# )


# class DatabricksModel(Model):
#     provider_name: str = Provider.databricks.value

#     supports_vision: bool = False


# DBU_USD_RATE = 0.07

# MODELS: list[Model] = [
#     DatabricksModel(
#         name="databricks-meta-llama-3-1-70b-instruct",
#         cost=Cost(prompt=14.286 * DBU_USD_RATE, completion=42.857 * DBU_USD_RATE),
#         context_length=8000,
#         vendor_name=Vendor.meta_llama.value,
#     ),
#     DatabricksModel(
#         name="databricks-meta-llama-3-1-405b-instruct",
#         cost=Cost(prompt=71.429 * DBU_USD_RATE, completion=214.286 * DBU_USD_RATE),
#         context_length=128000,
#         vendor_name=Vendor.meta_llama.value,
#     ),
#     DatabricksModel(
#         name="databricks-mixtral-8x7b-instruct",
#         cost=Cost(prompt=7.143 * DBU_USD_RATE, completion=14.286 * DBU_USD_RATE),
#         context_length=32000,
#         vendor_name=Vendor.databricks.value,
#         supports_tools=False,
#     ),
#     DatabricksModel(
#         name="databricks-dbrx-instruct",
#         cost=Cost(prompt=10.714 * DBU_USD_RATE, completion=32.143 * DBU_USD_RATE),
#         context_length=32000,
#         completion_length=4000,
#         vendor_name=Vendor.databricks.value,
#     ),
# ]


# class DatabricksSDKChatProviderAdapter(OpenAISDKChatAdapter):
#     @staticmethod
#     def get_supported_models() -> list[Model]:
#         return MODELS

#     @staticmethod
#     def get_api_key_name() -> str:
#         return "DATABRICKS_API_KEY"

#     def get_base_sdk_url(self) -> str:
#         return "https://adb-8736858266948228.8.azuredatabricks.net/serving-endpoints"

#     def _get_params(
#         self,
#         llm_input: Conversation,
#         **kwargs: Any,
#     ) -> Dict[str, Any]:
#         params = super()._get_params(llm_input, **kwargs)

#         messages = params["messages"]
#         databricksTools = kwargs.get("tools")

#         # Databricks only support system as a first optional message
#         if messages and messages[0]["role"] == ConversationRole.system:
#             system_message = messages[0]
#             messages = [system_message] + [
#                 msg for msg in messages[1:] if msg["role"] != ConversationRole.system
#             ]
#         else:
#             messages = [
#                 msg for msg in messages if msg["role"] != ConversationRole.system
#             ]

#         # Databricks only support ending messages with user or tool roles
#         if messages and messages[-1]["role"] not in [
#             ConversationRole.user,
#             ConversationRole.tool,
#         ]:
#             messages = messages + [{"role": ConversationRole.user, "content": ""}]

#         if databricksTools and not databricksTools[0]["function"].get("parameters"):
#             databricksTools[0]["function"]["parameters"] = {
#                 "type": "object",
#             }

#         return {
#             **params,
#             "messages": messages,
#             "tools": databricksTools,
#             "max_tokens": (
#                 kwargs.get("max_tokens")
#                 if kwargs.get("max_tokens")
#                 else self.get_model().completion_length
#             ),
#         }



import os

OVERRIDE_ALL_BASE_URLS = os.getenv("_ADAPTERS_OVERRIDE_ALL_BASE_URLS_")

# Has fallback env for backwards compatibility
MAX_CONNECTIONS_PER_PROCESS = int(
    os.getenv(
        "ADAPTERS_MAX_CONNECTIONS_PER_PROCESS",
        os.getenv("MAX_CONNECTIONS_PER_PROCESS", "1000"),
    )
)
MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS = int(
    os.getenv(
        "ADAPTERS_MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS",
        os.getenv("MAX_KEEPALIVE_CONNECTIONS_PER_PROCESS", "100"),
    )
)
HTTP_TIMEOUT = float(
    os.getenv("ADAPTERS_HTTP_TIMEOUT", os.getenv("HTTP_TIMEOUT", "600.0"))
)
HTTP_CONNECT_TIMEOUT = float(
    os.getenv("ADAPTERS_HTTP_CONNECT_TIMEOUT", os.getenv("HTTP_CONNECT_TIMEOUT", "5.0"))
)



from enum import Enum
import time
from typing import Any, Dict

from cohere import (
    AsyncClientV2,
    ChatContentDeltaEventDelta,
    ChatContentDeltaEventDeltaMessage,
    ChatContentDeltaEventDeltaMessageContent,
    ChatMessageEndEventDelta,
    ChatResponse,
    ClientV2,
    ContentDeltaStreamedChatResponseV2,
    MessageEndStreamedChatResponseV2,
    MessageStartStreamedChatResponseV2,
    StreamedChatResponseV2,
)

from llm_adapters.abstract_adapters.sdk_chat_adapter import SDKChatAdapter
from llm_adapters.types import (
    CompletionUsage,
    ChatCompletionMessage,
    Choice,
    ChoiceChunk,
    ChoiceDelta,
    AdapterChatCompletion,
    AdapterChatCompletionChunk,
    AdapterFinishReason,
    ConversationRole,
    Model,
    Provider,
    Vendor,
    # ChatCompletionMessageParam,
)
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)


class CohereModel(Model):
    provider_name: str = Provider.cohere.value
    vendor_name: str = Vendor.cohere.value

    supports_tools: bool = False
    supports_functions: bool = False
    supports_vision: bool = False
    supports_completion: bool = False
    supports_json_output: bool = False
    supports_n: bool = False
    supports_stop: bool = False

    can_empty_content: bool = False
    can_system_only: bool = False
    can_min_p: bool = False
    can_top_p: bool = False
    can_top_k: bool = False


# TODO: check completion length
MODELS: list[Model] = [
    # CohereModel(
    #     name="command",
    #     cost=Cost(prompt=1.00e-6, completion=2.00e-6),
    #     context_length=4000,
    #     completion_length=4000,
    # ),
    # CohereModel(
    #     name="command-light",
    #     cost=Cost(prompt=0.30e-6, completion=0.60e-6),
    #     context_length=4000,
    #     completion_length=4000,
    # ),
    # CohereModel(
    #     name="command-r",
    #     cost=Cost(prompt=0.15e-6, completion=0.60e-6),
    #     context_length=128000,
    #     completion_length=4000,
    # ),
    # CohereModel(
    #     name="command-r-plus",
    #     cost=Cost(prompt=2.50e-6, completion=10.00e-6),
    #     context_length=128000,
    #     completion_length=4000,
    # ),
    # CohereModel(
    #     name="command-r7b-12-2024",
    #     cost=Cost(prompt=0.0375e-6, completion=0.15e-6),
    #     context_length=128000,
    #     completion_length=4000,
    # ),
    # CohereModel(
    #     name="c4ai-aya-expanse-8b",
    #     cost=Cost(prompt=0.50e-6, completion=1.50e-6),
    #     context_length=8000,
    #     completion_length=4000,
    # ),
    # CohereModel(
    #     name="c4ai-aya-expanse-32b",
    #     cost=Cost(prompt=0.50e-6, completion=1.50e-6),
    #     context_length=128000,
    #     completion_length=4000,
    # ),
]


class CohereFinishReason(str, Enum):
    complete = "COMPLETE"
    max_tokens = "MAX_TOKENS"
    stop_sequence = "STOP_SEQUENCE"
    tool_call = "TOOL_CALL"
    error = "ERROR"


FINISH_REASON_MAPPING: Dict[CohereFinishReason, AdapterFinishReason] = {
    CohereFinishReason.complete: AdapterFinishReason.stop,
    CohereFinishReason.max_tokens: AdapterFinishReason.length,
    CohereFinishReason.stop_sequence: AdapterFinishReason.stop,
    CohereFinishReason.tool_call: AdapterFinishReason.tool_calls,
}


class CohereSDKChatProviderAdapter(SDKChatAdapter[ClientV2, AsyncClientV2]):
    @staticmethod
    def get_supported_models() -> list[Model]:
        return MODELS

    @staticmethod
    def get_api_key_name() -> str:
        return "COHERE_API_KEY"

    def get_base_sdk_url(self) -> str:
        return "https://api.cohere.com"

    def _sync_client_wrapper(self, **kwargs: Any) -> Any:
        if kwargs.pop("stream", False):
            return self._client_sync.chat_stream(**kwargs)
        return self._client_sync.chat(**kwargs)

    async def _async_client_wrapper(self, **kwargs: Any) -> Any:
        if kwargs.pop("stream", False):
            return self._client_async.chat_stream(**kwargs)
        return await self._client_async.chat(**kwargs)

    def _call_async(self) -> Any:
        return self._async_client_wrapper

    def _call_sync(self) -> Any:
        return self._sync_client_wrapper

    def _call_completion_sync(self) -> Any:
        raise NotImplementedError

    def _call_completion_async(self) -> Any:
        raise NotImplementedError

    def _create_client_sync(self, base_url: str, api_key: str) -> ClientV2:
        return ClientV2(base_url=base_url, api_key=api_key)  # type: ignore

    def _create_client_async(self, base_url: str, api_key: str) -> AsyncClientV2:
        return AsyncClientV2(base_url=base_url, api_key=api_key)  # type: ignore

    def _adjust_temperature(self, temperature: float) -> float:
        return temperature / 2

    # def _get_params(
    #     self, messages: list[ChatCompletionMessageParam], **kwargs: Any
    # ) -> dict[str, Any]:
    #     params = super()._get_params(messages, **kwargs)

    #     # Cohere does not support extra_body
    #     params = {**params, **params["extra_body"]}
    #     del params["extra_body"]

    #     return params

    def _extract_response(
        self, request: Any, response: ChatResponse
    ) -> AdapterChatCompletion:
        prompt_tokens = int(
            response.usage.billed_units.input_tokens
            if response.usage
            and response.usage.billed_units
            and response.usage.billed_units.input_tokens
            else 0
        )
        completion_tokens = int(
            response.usage.billed_units.output_tokens
            if response.usage
            and response.usage.billed_units
            and response.usage.billed_units.output_tokens
            else 0
        )

        cost = (
            self.get_model().cost.prompt * prompt_tokens
            + self.get_model().cost.completion * completion_tokens
            + self.get_model().cost.request
        )

        finish_reason = FINISH_REASON_MAPPING.get(
            CohereFinishReason(response.finish_reason), AdapterFinishReason.stop
        )

        choices: list[Choice] = []
        if response.message and response.message.content:
            for content in response.message.content:
                choices.append(
                    Choice(
                        index=len(choices),
                        finish_reason=finish_reason.value,
                        message=ChatCompletionMessage(
                            role=ConversationRole.assistant.value,
                            content=content.text,
                        ),
                    )
                )
        elif response.message and response.message.tool_calls:
            if response.message.tool_plan:
                choices.append(
                    Choice(
                        index=len(choices),
                        finish_reason=finish_reason.value,
                        message=ChatCompletionMessage(
                            role=ConversationRole.assistant.value,
                            content=response.message.tool_plan,
                        ),
                    )
                )

            for tool_call in response.message.tool_calls:
                if (
                    not tool_call.id
                    or not tool_call.function
                    or not tool_call.function.name
                    or not tool_call.function.arguments
                ):
                    raise ValueError("Unsupported response")

                choices.append(
                    Choice(
                        index=len(choices),
                        finish_reason=finish_reason.value,
                        message=ChatCompletionMessage(
                            role=ConversationRole.assistant.value,
                            tool_calls=[
                                ChatCompletionMessageToolCall(
                                    id=tool_call.id,
                                    type="function",
                                    function=Function(
                                        name=tool_call.function.name,
                                        arguments=tool_call.function.arguments,
                                    ),
                                )
                            ],
                        ),
                    )
                )

        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )

        return AdapterChatCompletion(
            id=response.id,
            created=int(time.time()),
            model=self.get_model().name,
            object="chat.completion",
            cost=cost,
            usage=usage,
            choices=choices,
        )

    def _extract_stream_response(
        self, request: Any, response: StreamedChatResponseV2, state: dict[str, Any]
    ) -> AdapterChatCompletionChunk:
        choice_chunk = ChoiceChunk(
            index=0,
            delta=ChoiceDelta(role=ConversationRole.assistant.value, content=""),
        )

        # TODO: Add citation support
        if isinstance(response, MessageStartStreamedChatResponseV2):
            state["id"] = response.id
            state["created"] = int(time.time())
        elif (
            isinstance(response, ContentDeltaStreamedChatResponseV2)
            and isinstance(response.delta, ChatContentDeltaEventDelta)
            and isinstance(response.delta.message, ChatContentDeltaEventDeltaMessage)
            and isinstance(
                response.delta.message.content, ChatContentDeltaEventDeltaMessageContent
            )
            and isinstance(response.delta.message.content.text, str)
        ):
            choice_chunk.delta.content = response.delta.message.content.text
        elif isinstance(response, MessageEndStreamedChatResponseV2):
            choice_chunk.finish_reason = AdapterFinishReason.stop.value

            if isinstance(response.delta, ChatMessageEndEventDelta) and isinstance(
                response.delta.finish_reason, ChatMessageEndEventDelta
            ):
                choice_chunk.finish_reason = FINISH_REASON_MAPPING.get(
                    CohereFinishReason(response.delta.finish_reason),
                    AdapterFinishReason.stop,
                ).value
        # else:
        # raise ValueError("Unsupported response")

        return AdapterChatCompletionChunk(
            id=state["id"],
            choices=[choice_chunk],
            created=state["created"],
            model=self.get_model().name,
            object="chat.completion.chunk",
        )

    def _extract_completion_response(
        self,
        request: Any,
        response: Any,
    ) -> Any:
        raise NotImplementedError

    def _extract_completion_stream_response(
        self, request: Any, response: Any, state: dict[str, Any]
    ) -> Any:
        raise NotImplementedError
